#!/usr/bin/env python3

#=======================================================================
# Authors: Ben Woodcroft
#
# Unit tests.
#
# Copyright
#
# This is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License.
# If not, see <http://www.gnu.org/licenses/>.
#=======================================================================

import unittest
import os.path
import sys
import math

import extern
import json
import pandas as pd
import tempfile
from io import StringIO

from bird_tool_utils import in_tempdir

sys.path = [os.path.join(os.path.dirname(os.path.realpath(__file__)),'..')]+sys.path
sys.path = [os.path.join(os.path.dirname(os.path.realpath(__file__)),'..','..')]+sys.path
kingfisher = os.path.join(os.path.dirname(os.path.realpath(__file__)),'..','bin','kingfisher')
path_to_data = os.path.abspath(os.path.join(os.path.dirname(__file__), 'data'))

eg_df = pd.read_csv(StringIO('run,bioproject,Gbp,library_strategy,library_selection,model,sample_name,taxon_name\n'
    'SRR13774710,PRJNA630999,10.342,WGS,RANDOM,Illumina NovaSeq 6000,SCB2WXA,human gut metagenome'))

class Tests(unittest.TestCase):
    maxDiff = None
    
    def test_one_sample_annotate(self):
        self.assertEqual(
            'run        | bioproject | Gbp   | library_strategy | library_selection | model               | sample_name | taxon_name\n' \
            '---------- | ---------- | ----- | ---------------- | ----------------- | ------------------- | ----------- | ----------\n' \
            'ERR1739691 | PRJEB15706 | 2.382 | WGS              | RANDOM            | Illumina HiSeq 2500 | MM1_1       | metagenome\n',
            extern.run('{} annotate -r ERR1739691 --debug'.format(kingfisher)))

    def test_one_project_annotate(self):
        self.assertEqual('run        | bioproject | Gbp   | library_strategy | library_selection | model               | sample_name | taxon_name\n' \
        '---------- | ---------- | ----- | ---------------- | ----------------- | ------------------- | ----------- | ----------\n' \
        'ERR1739691 | PRJEB15706 | 2.382 | WGS              | RANDOM            | Illumina HiSeq 2500 | MM1_1       | metagenome\n' \
        'ERR1739692 | PRJEB15706 | 2.382 | WGS              | RANDOM            | Illumina HiSeq 2500 | MM1_2       | metagenome\n' \
        'ERR1739693 | PRJEB15706 | 2.364 | WGS              | RANDOM            | Illumina HiSeq 2500 | MM1_3       | metagenome\n' \
        'ERR1739694 | PRJEB15706 | 2.501 | WGS              | RANDOM            | Illumina HiSeq 2500 | MM1_4       | metagenome\n' \
        'ERR1739695 | PRJEB15706 | 2.379 | WGS              | RANDOM            | Illumina HiSeq 2500 | MM1_5       | metagenome\n' \
        'ERR1739696 | PRJEB15706 | 2.351 | WGS              | RANDOM            | Illumina HiSeq 2500 | MM1_6       | metagenome\n' \
        'ERR1739697 | PRJEB15706 | 2.524 | WGS              | RANDOM            | Illumina HiSeq 2500 | MM1_7       | metagenome\n' \
        'ERR1739698 | PRJEB15706 | 2.358 | WGS              | RANDOM            | Illumina HiSeq 2500 | MM1_8       | metagenome\n' \
        'ERR1739699 | PRJEB15706 | 2.465 | WGS              | RANDOM            | Illumina HiSeq 2500 | MM1_9       | metagenome\n',
            extern.run('{} annotate -p PRJEB15706 --debug'.format(kingfisher)))

    def test_one_sample_annotate_csv(self):
        self.assertEqual('run,bioproject,Gbp,library_strategy,library_selection,model,sample_name,taxon_name,experiment_accession,experiment_title,library_name,library_source,library_layout,platform,submitter,study_accession,study_alias,study_centre_project_name,organisation,organisation_department,organisation_institution,organisation_street,organisation_city,organisation_country,organisation_contact_name,organisation_contact_email,sample_description,sample_alias,sample_accession,biosample,sample_title,ENA first public,ENA last update,External Id,INSDC center alias,INSDC center name,INSDC first public,INSDC last update,INSDC status,Submitter Id,collection date,depth,environment (biome),environment (feature),environment (material),geographic location (country and/or sea),geographic location (depth),geographic location (elevation),geographic location (latitude),geographic location (longitude),investigation type,microbial mat/biofilm environmental package,project name,sample name,sample storage duration,sample storage temperature,sequencing method,study_title,design_description,study_abstract,study_links,number_of_runs_for_sample,spots,bases,run_size,published,read1_length_average,read1_length_stdev,read2_length_average,read2_length_stdev\n' \
            'ERR1739691,PRJEB15706,2.382,WGS,RANDOM,Illumina HiSeq 2500,MM1_1,metagenome,ERX1809317,Illumina HiSeq 2500 paired end sequencing,unspecified,METAGENOMIC,PAIRED,ILLUMINA,European Nucleotide Archive,ERP017539,ena-STUDY-NIOZ-10-10-2016-11:18:17:022-1157,Minimal Mat 1,Royal Netherlands Institute for Sea Research,,,,,, ,,"artificial minimal coastal microbial mats at dilution 0, replicate 1",SAMEA4497179,ERS1396358,SAMEA4497179,artificial minimal coastal microbial mats,2017-06-08,2016-11-23,SAMEA4497179,NIOZ,Royal Netherlands Institute for Sea Research,2017-06-08T17:01:18Z,2016-11-23T11:15:32Z,public,MM1_1,2015-11,0.01,Microbial Mat Material,Beach,soil,Netherlands,0,0,53.489606,6.139913,metagenome,microbial mat/biofilm,Minimal Mat,MM1_1,10,20,illumina PE100,construction of minimal coastal microbial mats,,"Minimal coastal microbial mats were created with diluted coastal mat samples obtained from the Dutch barrier island of Schiermonnikoog. The MM\'s were inoculated in fresh sterilized sand in glass containers contained in a MicroBox. The MicroBox has a transparent lid (allowing photosynthetic growth) and a gas exchange filter. The MM\'s are propagated under laboratory conditions at a 16h light / 8h dark regime and at a constant 23 C. Serial dilutions used for this data-set are 0, 3 and 5-fold.",[],1,7938968,2381690400,936643449,2017-06-13 08:05:22,150,0,150,0\n',
            extern.run('{} annotate -r ERR1739691 -f csv --all-columns'.format(kingfisher)))

    def test_json_to_stdout(self):
        self.assertEqual(
            [{"run":"SRR13774710","bioproject":"PRJNA630999","Gbp":10.342,"library_strategy":"WGS","library_selection":"RANDOM","model":"Illumina NovaSeq 6000","sample_name":"SCB2WXA","taxon_name":"human gut metagenome"}],
            json.loads(extern.run('{} annotate -r SRR13774710 --output-format json'.format(kingfisher))))

    def test_json_to_file(self):
        with tempfile.NamedTemporaryFile() as f:
            extern.run('{} annotate -r SRR13774710 --output-format json --output-file {}'.format(kingfisher, f.name))
            self.assertEqual(
                [{"run":"SRR13774710","bioproject":"PRJNA630999","Gbp":10.342,"library_strategy":"WGS","library_selection":"RANDOM","model":"Illumina NovaSeq 6000","sample_name":"SCB2WXA","taxon_name":"human gut metagenome"}],
                json.loads(f.read()))
    
    def test_parquet(self):
        with tempfile.NamedTemporaryFile() as f:
            extern.run('{} annotate -r SRR13774710 --output-format parquet --output-file {}'.format(kingfisher, f.name))
            self.assertEqual(eg_df.to_dict(), pd.read_parquet(f.name).to_dict())
    
    def test_parquet_all_columns(self):
        with tempfile.NamedTemporaryFile() as f:
            extern.run('{} annotate -r SRR13774710 SRR7051324 --all-columns --output-format parquet --output-file {}'.format(kingfisher, f.name))
            self.assertEqual(str(pd.read_parquet('{}/2_accessions.annotate.pq'.format(path_to_data)).to_dict()), str(pd.read_parquet(f.name).to_dict()))

    def test_feather(self):
        with tempfile.NamedTemporaryFile() as f:
            extern.run('{} annotate -r SRR13774710 --output-format feather --output-file {}'.format(kingfisher, f.name))
            self.assertEqual(eg_df.to_dict(), pd.read_feather(f.name).to_dict())

    def test_bases_missing_field(self):
        with tempfile.NamedTemporaryFile() as f:
            extern.run('{} annotate -r ERR2178284 --output-format csv --output-file {}'.format(kingfisher, f.name))
            expected = {'run': {0: 'ERR2178284'},
                'bioproject': {0: 'PRJEB23079'},
                'library_strategy': {0: 'WGS'},
                'library_selection': {0: 'Hybrid Selection'},
                'model': {0: 'Illumina HiSeq 2500'},
                'sample_name': {0: 'STR486'},
                'taxon_name': {0: 'Homo sapiens'}}
            observed = pd.read_csv(f.name).to_dict()
            self.assertTrue(math.isnan(observed['Gbp'][0]))
            del observed['Gbp'] # nan != nan apparently
            self.assertEqual(expected, observed)

    def test_multiple_bioprojects(self):
        expected = """run         | bioproject  | Gbp    | library_strategy | library_selection | model                 | sample_name | taxon_name        
----------- | ----------- | ------ | ---------------- | ----------------- | --------------------- | ----------- | ------------------
SRR11703773 | PRJNA621515 | 15.264 | WGS              | RANDOM            | Illumina NovaSeq 6000 | GNNTP       | wetland metagenome
SRR11703788 | PRJNA621514 | 21.045 | WGS              | RANDOM            | Illumina NovaSeq 6000 | GNNTO       | wetland metagenome
"""
        stdout = extern.run('{} annotate --bioprojects PRJNA621515 PRJNA621514'.format(kingfisher))
        self.assertEqual(stdout, expected)

    def test_api_stability(self):
        from kingfisher import annotate
        with in_tempdir():
            # https://github.com/NIAID-Data-Ecosystem/nde-crawlers/blob/c67c77d0c547466cc2d2264f8e53e0c809b9dfd0/ncbi_sra/files/ncbi_sra.py#L34
            annotate(
                run_identifiers=None,
                run_identifiers_file=None,
                bioproject_accession='PRJNA621515',
                output_file=f"PRJNA621515.json",
                output_format="json",
                all_columns=True,
            )
            with open('PRJNA621515.json') as f:
                j = json.load(f)
            self.assertEqual('SRR11703773', j[0]['run'])


if __name__ == "__main__":
    unittest.main()



import os
import time
import requests
import xml.etree.ElementTree as ET
import logging
import collections
import json
from tqdm import tqdm

try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO

import pandas as pd

from bird_tool_utils import iterable_chunks

# Define these constants so that they can be referred to in other classes
# without index errors.
STUDY_ACCESSION_KEY = 'study_accession'
RUN_ACCESSION_KEY = 'run'
BASES_KEY = 'bases'
SAMPLE_NAME_KEY = 'sample_name'
NCBI_API_KEY_ENV = 'NCBI_API_KEY'
BIOPROJECT_ACCESSION_KEY = 'bioproject'

class SraMetadata:
    def add_api_key(self, other_params):
        if NCBI_API_KEY_ENV in os.environ:
            other_params['api_key'] = os.environ[NCBI_API_KEY_ENV]
        return other_params

    def _retry_request(self, description, func):
        '''Retry a reqests.post or requests.get 3 times, returning the request
        when OK, otherwise raising an Exception'''

        num_retries = 3
        sleep_time = 60
        def retrying(i, num_retries=3):
            if i < num_retries-1:
                logging.warning("Retrying request (retry {} of {})".format(i+1, num_retries-1))
        
        for i in range(num_retries):
            try:
                this_res = func()
                if not this_res.ok:
                    logging.warning("Request not OK when {}: {}: {}".format(description, this_res, this_res.text))
                    logging.warning("Sleeping for {} seconds before retrying".format(sleep_time))
                    time.sleep(60)
                    retrying(i)
                else:
                    return this_res
            except Exception as e:
                logging.warning("Exception raised when {}: {}".format(description, e))
                logging.warning("Sleeping for {} seconds before retrying".format(sleep_time))
                time.sleep(60)
                retrying(i)
        raise Exception("Failed to {} after {} attempts".format(description, num_retries))


    def fetch_runs_from_bioprojects(self, bioproject_accessions):
        retmax = 10000
        query_string = " OR ".join(["{}[BioProject]".format(bioproject_accession) for bioproject_accession in bioproject_accessions])
        logging.debug("Querying with string: {}".format(query_string))
        res = requests.get(
            url="https://eutils.ncbi.nlm.nih.gov/entrez/eutils/esearch.fcgi",
            params=self.add_api_key({
                "db": "sra",
                "term": query_string,
                "tool": "kingfisher",
                "email": "kingfisher@github.com",
                "retmax": retmax,
                "usehistory": "y",
                }),
            )
        if not res.ok:
            raise Exception("HTTP Failure when requesting search from bioproject: {}: {}".format(res, res.text))
        root = ET.fromstring(res.text)
        sra_ids = list([c.text for c in root.find('IdList')])
        if len(sra_ids) == retmax:
            logging.warning("Unexpectedly found the maximum number of results for this query, possibly some results will be missing")
        webenv = root.find('WebEnv').text

        # Now convert the IDs into runs
        metadata = self.efetch_metadata_from_ids(webenv, None, len(sra_ids))
        return metadata[RUN_ACCESSION_KEY].to_list()

    def efetch_metadata_from_ids(self, webenv, accessions, num_ids):
        data_frames = []

        retmax = num_ids+10
        logging.debug("Running efetch ..")
        res = self._retry_request(
            'efetch_from_ids',
            lambda: requests.get(
                url="https://eutils.ncbi.nlm.nih.gov/entrez/eutils/efetch.fcgi",
                params=self.add_api_key({
                    "db": "sra",
                    "tool": "kingfisher",
                    "email": "kingfisher@github.com",
                    "webenv": webenv,
                    "query_key": 1
                    }),
                ))
        if not res.ok:
            raise Exception("HTTP Failure when requesting efetch from IDs: {}: {}".format(res, res.text))

        root = ET.fromstring(res.text)

        def try_get(func):
            try:
                return func()
            except AttributeError:
                return ''
            except KeyError:
                return None

        if root.find("ERROR") is not None:
            logging.error("Error when fetching metadata: {}".format(root.find("ERROR").text))

        # Some samples such as SAMN13241871 are linked to multiple runs e.g. SRR10489833
        accessions_set = None if accessions is None else set(accessions)

        for pkg in root.findall('EXPERIMENT_PACKAGE'):
            d = collections.OrderedDict()
            d['experiment_accession'] = try_get(lambda: pkg.find('./EXPERIMENT').attrib['accession'])
            d['experiment_title'] = try_get(lambda: pkg.find('./EXPERIMENT/TITLE').text)
            l = pkg.find('./EXPERIMENT/DESIGN/LIBRARY_DESCRIPTOR')
            d['library_name'] = try_get(lambda: l.find('LIBRARY_NAME').text)
            d['library_strategy'] = try_get(lambda: l.find('LIBRARY_STRATEGY').text)
            d['library_source'] = try_get(lambda: l.find('LIBRARY_SOURCE').text)
            d['library_selection'] = try_get(lambda: l.find('LIBRARY_SELECTION').text)
            d['library_layout'] = try_get(lambda: l.find('LIBRARY_LAYOUT')[0].tag)
            d['platform'] = try_get(lambda: pkg.find('./EXPERIMENT/PLATFORM')[0].tag)
            d['model'] = try_get(lambda: pkg.find('./EXPERIMENT/PLATFORM/')[0].text)
            d['submitter'] = ''
            for k, v in pkg.find('./SUBMISSION').attrib.items():
                if k not in ('accession','alias'):
                    if d['submitter'] == '':
                        d['submitter'] = v
                    else:
                        d['submitter'] = "{}, {}".format(d['submitter'], v)
            d[STUDY_ACCESSION_KEY] = try_get(lambda: pkg.find('./STUDY').attrib['accession'])
            d[BIOPROJECT_ACCESSION_KEY] = try_get(lambda: pkg.find('./STUDY/IDENTIFIERS/EXTERNAL_ID[@namespace="BioProject"]').text)
            d['study_alias'] = try_get(lambda: pkg.find('./STUDY').attrib['alias'])
            d['study_centre_project_name'] = try_get(lambda: pkg.find('./STUDY/DESCRIPTOR/CENTER_PROJECT_NAME').text)
            d['organisation'] = try_get(lambda: pkg.find('./Organization/Name').text)
            d['organisation_department'] = try_get(lambda: pkg.find('./Organization/Address/Department').text)
            d['organisation_institution'] = try_get(lambda: pkg.find('./Organization/Address/Institution').text)
            d['organisation_street'] = try_get(lambda: pkg.find('./Organization/Address/Street').text)
            d['organisation_city'] = try_get(lambda: pkg.find('./Organization/Address/City').text)
            d['organisation_country'] = try_get(lambda: pkg.find('./Organization/Address/Country').text)
            first_name = try_get(lambda: pkg.find('./Organization/Contact/Name/First').text)
            last_name = try_get(lambda: pkg.find('./Organization/Contact/Name/Last').text)
            d['organisation_contact_name'] = "{} {}".format(first_name, last_name)
            d['organisation_contact_email'] = try_get(lambda: pkg.find('./Organization/Contact').attrib['email'])
            # Remove carriage return from sample description e.g. for SRR1263047
            d['sample_description'] = try_get(lambda: pkg.find('./SAMPLE/DESCRIPTION').text.replace('\r',''))
            d['sample_alias'] = try_get(lambda: pkg.find('./SAMPLE').attrib['alias'])
            d['sample_accession'] = try_get(lambda: pkg.find('./SAMPLE').attrib['accession'])
            d['biosample'] = try_get(lambda: pkg.find('./SAMPLE/IDENTIFIERS/EXTERNAL_ID[@namespace="BioSample"]').text)
            d['sample_title'] = try_get(lambda: pkg.find('./SAMPLE/TITLE').text)
            d['taxon_name'] = try_get(lambda: pkg.find('./SAMPLE/SAMPLE_NAME/SCIENTIFIC_NAME').text)
            sample_sample_name = None
            sample_title = None
            if pkg.find('./SAMPLE/SAMPLE_ATTRIBUTES'):
                for attr in pkg.find('./SAMPLE/SAMPLE_ATTRIBUTES'):
                    tag_el = attr.find('TAG')
                    value_el = attr.find('VALUE')

                    # Some samples like
                    # https://www.ncbi.nlm.nih.gov/sra?term=ERS1240061&report=FullXml
                    # have entries with a tag, but no value. Ignore these.
                    if tag_el is not None and value_el is not None:
                        tag = tag_el.text
                        value = value_el.text
                        if tag == 'Title':
                            sample_title = value
                        elif tag == 'sample name':
                            sample_sample_name = value
                        d[tag] = value
            if sample_sample_name is not None:
                d[SAMPLE_NAME_KEY] = sample_sample_name
            elif sample_title is not None:
                d[SAMPLE_NAME_KEY] = sample_title
            else:
                d[SAMPLE_NAME_KEY] = d['library_name'] #default, maybe there's always a title though?
            d['study_title'] = try_get(lambda: pkg.find('./STUDY/DESCRIPTOR/STUDY_TITLE').text)
            d['design_description'] = try_get(lambda: pkg.find('./EXPERIMENT/DESIGN/DESIGN_DESCRIPTION').text)
            d['study_abstract'] = try_get(lambda: pkg.find('./STUDY/DESCRIPTOR/STUDY_ABSTRACT').text)
            study_links_xrefs = try_get(lambda: pkg.find('./STUDY/STUDY_LINKS'))
            if study_links_xrefs is not None:
                # Convert db to lower case because otherwise have PUBMED and pubmed e.g. ERR1914274 and SRR9113719
                study_links = list([
                    {'db': x.find('DB').text.lower(), 'id': x.find('ID').text}
                    for x in study_links_xrefs.findall('STUDY_LINK/XREF_LINK')])
                # Record URL links like SRR7051324
                for x in study_links_xrefs.findall('STUDY_LINK/URL_LINK'):
                    study_links.append({
                        'label': x.find('LABEL').text,
                        'url': x.find('URL').text
                    })
                d['study_links'] = json.dumps(study_links)
            else:
                d['study_links'] = json.dumps([])
            
            # Account for the fact that multiple runs may be associated with
            # this sample
            d['number_of_runs_for_sample'] = len(pkg.findall('./RUN_SET/RUN'))

            for run in pkg.findall('./RUN_SET/RUN'):
                accession_here = run.attrib['accession']
                if accessions_set is None or accession_here in accessions_set:
                    d2 = d.copy()
                    d2['spots'] = try_get(lambda: int(run.attrib['total_spots']))
                    d2[BASES_KEY] = try_get(lambda: int(run.attrib['total_bases']))
                    d2['run_size'] = try_get(lambda: int(run.attrib['size']))
                    d2[RUN_ACCESSION_KEY] = try_get(lambda: run.attrib['accession'])
                    d2['published'] = try_get(lambda: run.attrib['published'])
                    stats = run.find('Statistics')
                    if stats is not None:
                        for (i, r) in enumerate(stats):
                            d2['read{}_length_average'.format(i+1)] = r.attrib['average']
                            d2['read{}_length_stdev'.format(i+1)] = r.attrib['stdev']
                    data_frames.append(d2)

        return pd.DataFrame(data_frames)

    def _print_xml(self, element, prefix):
        if prefix is None or prefix == '':
            p2 = ""
        else:
            p2 = "{}_".format(prefix)
        for k, v in element.attrib.items():
            print("\t".join(['{}{}'.format(p2,k),v]))
        if element.text:
            print("\t".join(['{}{}'.format(p2, element.tag), element.text]))
        for e in element:
            self.print_xml(e, '{}{}'.format(p2, e.tag))

    def efetch_sra_from_accessions(self, accessions):
        all_accessions = list(set(accessions))
        if len(all_accessions) == 0:
            return []

        metadata_chunks = []
        chunk_size = 500

        # Complicated calculation here.
        num_chunks = len(all_accessions)/chunk_size
        if num_chunks > int(num_chunks):
            num_chunks = int(num_chunks) + 1
        num_chunks = int(num_chunks)

        if len(all_accessions) > chunk_size:
            logging.info("Querying for {} accessions in {} chunks".format(len(all_accessions), num_chunks))
            accession_iter = tqdm(iterable_chunks(all_accessions, chunk_size), total=num_chunks)
        else:
            accession_iter = [all_accessions]

        for accessions_plus in accession_iter:
            accessions = [a for a in accessions_plus if a is not None]

            if num_chunks == 1:
                logging.info("Querying NCBI esearch for {} distinct accessions e.g. {}".format(
                    len(accessions), accessions[0]))
            sra_ids = []

            webenv = None
            request_term = ' OR '.join(["{}[accn]".format(acc) for acc in accessions])

            retmax = len(accessions)+10
            params=self.add_api_key({
                "db": "sra",
                "term": request_term,
                "tool": "kingfisher",
                "email": "kingfisher@github.com",
                "retmax": retmax,
                "usehistory": "y",
                })
            if webenv is None:
                params['WebEnv'] = webenv

            res = self._retry_request(
                "esearch from accessions", 
                lambda: requests.post(
                    url="https://eutils.ncbi.nlm.nih.gov/entrez/eutils/esearch.fcgi",
                    data=params))

            root = ET.fromstring(res.text)
            if webenv is None:
                webenv = root.find('WebEnv').text
            id_list_node = root.find('IdList')
            sra_ids = list(set([c.text for c in id_list_node]))

            if len(sra_ids) == 0:
                logging.warning("Unable to find any accessions, from the list: {}".format(accessions))
                return None

            if num_chunks == 1:
                logging.info("Querying NCBI efetch for {} distinct IDs e.g. {}".format(
                    len(sra_ids), sra_ids[0]))
            metadata = self.efetch_metadata_from_ids(webenv, accessions, len(sra_ids))

            # Ensure all hits are found, and trim results to just those that are real hits
            if RUN_ACCESSION_KEY not in metadata.columns:
                raise Exception("No metadata could be retrieved")

            if len(metadata) != len(accessions):
                found_runs = set(metadata[RUN_ACCESSION_KEY].to_list())
                not_found = list([a for a in accessions if a not in found_runs])
                logging.warning("Unable to find all accessions. The {} missing ones were: {}".format(
                    len(not_found), not_found
                ))
            metadata_chunks.append(metadata)

        metadata = pd.concat(metadata_chunks)
        metadata.sort_values([STUDY_ACCESSION_KEY,RUN_ACCESSION_KEY], inplace=True)

        return metadata



from .version import __version__

import logging
import json
import os
import subprocess
import sys
import gzip
import re

import extern
from extern import ExternCalledProcessError
import bird_tool_utils

from .ena import EnaDownloader
from .location import Location, NcbiLocationJson
from .exception import DownloadMethodFailed
from .sra_metadata import *
from .md5sum import MD5

DEFAULT_ASPERA_SSH_KEY = 'linux'
DEFAULT_OUTPUT_FORMAT_POSSIBILITIES = ['fastq', 'fastq.gz']
DEFAULT_THREADS = 8
DEFAULT_DOWNLOAD_THREADS = DEFAULT_THREADS
DEFAULT_ASCP_ARGS = '-k 2'

class OutputLocation:
    def __init__(self, output_directory):
        self.output_directory = os.path.abspath(output_directory)

        # Check to make sure the directory exists, or create it if is only 1 directory to be created.
        if not os.path.exists(self.output_directory):
            if not os.path.exists(os.path.dirname(self.output_directory)):
                raise Exception("The parent directory of the output directory specified '{}' does not exist. Kingfisher will create 1 directory if required, but not 2 or more. If you really want to output to that directory, you could try using 'mkdir -p '{}'".format(
                    self.output_directory,
                    self.output_directory))
            else:
                logging.info("Creating output directory {}".format(self.output_directory))
                os.mkdir(self.output_directory)

    def output_stem(self, run_identifier):
        return os.path.join(self.output_directory, run_identifier)


def download_and_extract(**kwargs):
    '''download an public sequence dataset and extract if necessary. kwargs
    here are largely the same as the arguments to the kingfisher executable.
    '''
    run_identifiers = kwargs.pop('run_identifiers')
    run_identifiers_file = kwargs.pop('run_identifiers_file')
    bioproject_accession = kwargs.pop('bioproject_accession', None)  # kept for API stability
    bioproject_accessions = kwargs.pop('bioproject_accessions', None)

    if bioproject_accession and bioproject_accessions is None:
        bioproject_accessions = [bioproject_accession]

    num_inputs = 0
    if run_identifiers is not None: num_inputs += 1
    if run_identifiers_file is not None: num_inputs += 1
    if bioproject_accessions is not None: num_inputs += 1
    if num_inputs != 1:
        raise Exception("Must specify exactly one input type: --run-identifiers, --bioproject-accessions or --run-identifiers-list")

    if bioproject_accessions is not None:
        run_identifiers = SraMetadata().fetch_runs_from_bioprojects(bioproject_accessions)
        logging.debug("Found {} run(s) to annotate".format(len(run_identifiers)))
    if run_identifiers_file is not None:
        with open(run_identifiers_file) as f:
            run_identifiers = list([r.strip() for r in f.readlines()])

    for run in run_identifiers:
        download_and_extract_one_run(run, **kwargs)

def download_and_extract_one_run(run_identifier, **kwargs):
    logging.debug("kwargs in download_and_extract_one_run: {}".format(kwargs))
    download_methods = kwargs.pop('download_methods')
    output_format_possibilities = kwargs.pop('output_format_possibilities',
        DEFAULT_OUTPUT_FORMAT_POSSIBILITIES)
    force = kwargs.pop('force', False)
    unsorted = kwargs.pop('unsorted', False)
    stdout = kwargs.pop('stdout', False)
    gcp_project = kwargs.pop('gcp_project', None)
    gcp_user_key_file = kwargs.pop('gcp_user_key_file', None)
    aws_user_key_id = kwargs.pop('aws_user_key_id', None)
    aws_user_key_secret = kwargs.pop('aws_user_key_secret', None)
    allow_paid = kwargs.pop('allow_paid', None)
    allow_paid_from_gcp = kwargs.pop('allow_paid_from_gcp', None)
    allow_paid_from_aws = kwargs.pop('allow_paid_from_aws', None)
    guess_aws_location = kwargs.pop('guess_aws_location', False)
    ascp_ssh_key = kwargs.pop('ascp_ssh_key', DEFAULT_ASPERA_SSH_KEY)
    ascp_args = kwargs.pop('ascp_args', DEFAULT_ASCP_ARGS)
    download_threads = kwargs.pop('download_threads', DEFAULT_DOWNLOAD_THREADS)
    extraction_threads = kwargs.pop('extraction_threads', DEFAULT_THREADS)
    hide_download_progress = kwargs.pop('hide_download_progress', False)
    prefetch_max_size = kwargs.pop('prefetch_max_size',None)
    check_md5sums = kwargs.pop('check_md5sums', False)
    output_directory = kwargs.pop('output_directory', '.')

    if len(kwargs) > 0:
        raise Exception("Unexpected arguments detected: %s" % kwargs)

    if guess_aws_location and check_md5sums:
        logging.warning("Guessing AWS location is not compatible with checking md5sums. Not carrying out md5sum checks for downloads from AWS.")

    if allow_paid:
        allowable_sources = ['s3', 'gcp']
    else:
        allowable_sources = []
    if allow_paid or allow_paid_from_gcp:
        if 'gcp-cp' not in download_methods:
            logging.warning("Allowing download from requester-pays GCP buckets, "\
                "but gcp-cp is not specified download method, so --allow-paid and --allow-paid-from-gcp have no effect")
        allowable_sources.append('gcp')
    if allow_paid_from_aws:
        if 'aws-cp' not in download_methods:
            logging.warning("Allowing download from requester-pays AWS buckets, "\
                "but aws-cp is not specified download method, so --allow-paid-from-aws has no effect")
        allowable_sources.append('s3')
        
    logging.debug("Allowing non-NCBI sources for download: {}".format(allowable_sources))

    if gcp_project and gcp_user_key_file:
        raise Exception("--gcp-project is incompatible with --gcp-user-key-file. The project specified in the key file will be used when gcp_project is not specified.")

    if stdout and not unsorted:
        raise Exception("Currently --stdout must be used with --unsorted")

    output_location_factory = OutputLocation(output_directory)
    output_files = []
    ncbi_locations = None

    # Checking for already existing files
    if stdout:
        skip_download_and_extraction, output_files = False, []
    else:
        skip_download_and_extraction, output_files = _check_for_existing_files(
            output_location_factory, run_identifier, output_format_possibilities, force
        )

    downloaded_files = None
    if not skip_download_and_extraction:
        # Download phase
        worked = False
        for method in download_methods:
            logging.info("Attempting download method {} for run {} ..".format(method, run_identifier))
            if method == 'prefetch':
                output_path = output_location_factory.output_stem('{}.sra'.format(run_identifier))
                try:
                    if prefetch_max_size is None:
                        prefetch_max_size_argument = '--max-size 0G'
                    else:
                        prefetch_max_size_argument = '--max-size {}'.format(prefetch_max_size)
                    extern.run("prefetch {} -o {} {}".format(
                        prefetch_max_size_argument, output_path, run_identifier))
                    if os.path.exists(output_path):
                        downloaded_files = [output_path]
                    else:
                        logging.warning("Method {} failed: Prefetch did not create file {}".format(method, output_path))
                except ExternCalledProcessError as e:
                    logging.warning("Method {} failed: Error was: {}".format(method, e))
                    if os.path.exists(output_path):
                        logging.info("Removing file {} because download failed ..".format(output_path))
                        os.remove(output_path)
                
            elif method == 'aws-http':
                def download_from_aws(odp_link, run_identifier, download_threads, method):
                    output_path = output_location_factory.output_stem('{}.sra'.format(run_identifier))
                    try:
                        if download_threads > 1:
                            logging.info(
                                "Downloading .SRA file from AWS Open Data Program HTTP link using aria2c ..")
                            verbosity_flag = '--quiet' if hide_download_progress else ''
                            # Redirect aria2c stdout to stderr so all logging of kingfisher is on stderr.
                            # aria2c does not handle absolute paths properly, so we have to use a relative path.
                            cmd = "aria2c {} -x{} -o {} '{}' 1>&2".format(
                                verbosity_flag, download_threads, os.path.relpath(output_path), odp_link)
                            subprocess.check_call(cmd, shell=True)
                        else:
                            logging.info(
                                "Downloading .SRA file from AWS Open Data Program HTTP link using curl ..")
                            verbosity_flag = '--silent --show-error' if hide_download_progress else ''
                            cmd = "curl {} -o {} '{}'".format(verbosity_flag, output_path, odp_link)
                            subprocess.check_call(cmd, shell=True)
                        logging.info("Download finished, validating ..")
                        # A download with curl of a bad AWS address does not
                        # result in a non-zero exitstatus. Instead an XML
                        # document is returned. If it is XML, then download has
                        # failed.
                        with open(output_path,'rb') as f:
                            aws_failed = (f.read(8) != b'NCBI.sra')

                        if aws_failed:
                            logging.info("The file downloaded from AWS appears not to be a .sra file, deleting it, this download method failed")
                            os.remove(output_path)
                            return None
                        else:
                            return [output_path]
                    except subprocess.CalledProcessError as e:
                        logging.warning("Method {} failed when downloading from {}: Error was: {}".format(method, odp_link, e))
                        if os.path.exists(output_path):
                            logging.info("Removing file {} because download failed ..".format(output_path))
                            os.remove(output_path)
                        return None

                if guess_aws_location:
                    # e.g. https://sra-pub-run-odp.s3.amazonaws.com/sra/SRR12118866/SRR12118866
                    guessed_location = 'https://sra-pub-run-odp.s3.amazonaws.com/sra/{}/{}'.format(run_identifier, run_identifier)
                    logging.info("Guessing AWS-ODP link to be: {}".format(guessed_location))
                    downloaded_files = download_from_aws(guessed_location, run_identifier, download_threads, method)
                else:
                    if ncbi_locations is None:
                        ncbi_locations = Location.get_ncbi_locations(run_identifier)
                    odp_http_locations = ncbi_locations.object_locations(
                        NcbiLocationJson.OBJECT_TYPE_SRA, NcbiLocationJson.AWS_SERVICE, False
                    )

                    if len(odp_http_locations) > 0:
                        for odp_http_location in odp_http_locations:
                            logging.debug("Found ODP link {}".format(odp_http_location))
                            logging.info("Found ODP link {}".format(odp_http_location.link()))
                            odp_link = odp_http_location.link()
                            downloaded_files = download_from_aws(odp_link, run_identifier, download_threads, method)
                            if downloaded_files is not None and check_md5sums:
                                for downloaded_file in downloaded_files:
                                    # Is there always just 1 .sra file? There is only 1 md5sum
                                    logging.info("Checking md5sum of downloaded file {} ..".format(downloaded_file))
                                    if MD5.check_md5sum(downloaded_file, odp_http_location.md5sum()):
                                        logging.info("MD5sum OK for {}".format(downloaded_file))
                                    else:
                                        logging.warning("MD5sum check failed for {}".format(downloaded_file))
                    else:
                        logging.warning("Method {} failed: No ODP URL could be found".format(method))

            elif method == 'aws-cp':
                if ncbi_locations is None:
                    ncbi_locations = Location.get_ncbi_locations(run_identifier)

                s3_locations = ncbi_locations.object_locations(
                    NcbiLocationJson.OBJECT_TYPE_SRA,
                    NcbiLocationJson.AWS_SERVICE,
                    's3' in allowable_sources
                )

                # TODO: Sort so unpaid are first
                output_path = output_location_factory.output_stem('{}.sra'.format(run_identifier))

                if len(s3_locations) > 0:
                    for s3_location in s3_locations:
                        logging.info("Found s3 link {}".format(s3_location.link()))

                        try:
                            command = '{} {}'.format(
                                s3_location.s3_command_prefix(run_identifier), output_path
                            )
                            if aws_user_key_id:
                                os.environ['AWS_ACCESS_KEY_ID'] = aws_user_key_id
                            if aws_user_key_id:
                                os.environ['AWS_SECRET_ACCESS_KEY'] = aws_user_key_secret
                            logging.info("Downloading from S3..")
                            try:
                                extern.run(command)
                                downloaded_files = [output_path]
                            except ExternCalledProcessError as e:
                                logging.warning("Method {} failed: Error was: {}".format(method, e))
                        except DownloadMethodFailed as e:
                            logging.warning("Method {} failed, error was {}".format(
                                method, e
                            ))
                            if os.path.exists(output_path):
                                logging.info("Removing file {} because download failed ..".format(output_path))
                                os.remove(output_path)
                else:
                    logging.warning("Method {} failed: No S3 location could be found".format(method))
                    if os.path.exists(output_path):
                        logging.info("Removing file {} because download failed ..".format(output_path))
                        os.remove(output_path)

            elif method == 'gcp-cp':
                output_path = output_location_factory.output_stem('{}.sra'.format(run_identifier))
                if 'gcp' in allowable_sources:
                    if ncbi_locations is None:
                        ncbi_locations = Location.get_ncbi_locations(run_identifier)
                    locations = ncbi_locations.object_locations(
                        NcbiLocationJson.OBJECT_TYPE_SRA, NcbiLocationJson.GCP_SERVICE, True
                    )
                    if len(locations) > 0:
                        for loc in locations:
                            command = 'gsutil'
                            gcp_project = gcp_project
                            if gcp_user_key_file:
                                with open(gcp_user_key_file) as f:
                                    j = json.load(f)
                                    if 'project_id' not in j:
                                        raise Exception("Unexpectedly could not find project_id in GCP user key JSON file")
                                    gcp_project = j['project_id']
                                extern.run('gcloud auth activate-service-account --key-file={}'.format(gcp_user_key_file))

                            failed = False
                            if gcp_project:
                                command = command + " -u {}".format(gcp_project)
                            else:
                                logging.info("Finding Google cloud project to charge")
                                project_id = extern.run('gcloud config get-value project').strip()
                                if project_id == '':
                                    logging.warning("Method gcp-cp failed: Could not find a GCP project to charge, cannot continue. "\
                                        "Expected a project from 'gcloud config get-value project' or specified with --gcp-user-key-file or --gcp-project")
                                    failed = True
                                else:
                                    logging.info("Charging to project \'{}\'".format(project_id))
                                    command = command + " -u {}".format(project_id)
                            if not failed:
                                try:
                                    gs_path = loc.gs_path()
                                    command += ' cp {} {}'.format(
                                        gs_path, output_path
                                    )
                                    logging.info("Downloading from GCP..")
                                    try:
                                        extern.run(command)
                                        downloaded_files = [output_path]
                                    except ExternCalledProcessError as e:
                                        logging.warning("Method {} failed: Error was: {}".format(method, e))
                                        
                                except DownloadMethodFailed as e:
                                    logging.warning("Method {} failed, error was {}".format(
                                        method, e
                                    ))
                                    if os.path.exists(output_path):
                                        logging.info("Removing file {} because download failed ..".format(output_path))
                                        os.remove(output_path)
                    else:
                        logging.warning("Method {} failed: No GCP location could be found".format(method))
                else:
                    logging.warning("Not using method gcp-cp as --allow-paid was not specified")

            elif method == 'ena-ascp':
                result = EnaDownloader().download_with_aspera(run_identifier, output_directory,
                    ascp_args=ascp_args,
                    ssh_key=ascp_ssh_key,
                    check_md5sums=check_md5sums)
                if result is not False:
                    gzip_test_files(result)
                    downloaded_files = result

            elif method == 'ena-ftp':
                result = EnaDownloader().download_with_curl(
                    run_identifier,
                    download_threads,
                    output_directory,
                    check_md5sums=check_md5sums)
                if result is not False:
                    gzip_test_files(result)
                    downloaded_files = result

            else:
                raise Exception("Unknown method: {}".format(method))
            
            if downloaded_files is not None:
                logging.info("Method {} worked.".format(method))
                break
            else:
                logging.warning("Method {} failed".format(method))

        if downloaded_files is None:
            raise Exception("No more specified download methods, cannot continue")

    # Extraction/conversion phase
    if not skip_download_and_extraction:
        if downloaded_files == [output_location_factory.output_stem('{}.sra'.format(run_identifier))]:
            sra_file = downloaded_files[0]
            if 'sra' not in output_format_possibilities:
                output_files = extract(
                    sra_file = sra_file,
                    output_format_possibilities = output_format_possibilities,
                    unsorted = unsorted,
                    stdout = stdout,
                    threads = extraction_threads,
                    output_directory = output_directory,
                )
                os.remove(sra_file)
            else:
                output_files.append(sra_file)
        else:
            if stdout:
                raise Exception("--stdout currently must be via download of a .sra format file, rather than a download from ENA. I imagine this will be fixed in future.")
            if 'fastq.gz' in output_format_possibilities:
                output_files = downloaded_files
            else:
                for fq in ['x_1.fastq.gz','x_2.fastq.gz','x.fastq.gz']:
                    f = output_location_factory.output_stem(fq.replace('x',run_identifier))
                    if os.path.exists(f):
                        # Do the least work, currently we have FASTQ.gz
                        if 'fasta' in output_format_possibilities:
                            logging.info("Converting {} to FASTA ..".format(f))
                            out_here = f.replace('.fastq.gz','.fasta')
                            extern.run("pigz -p {} -cd {} |awk '{{print \">\" substr($0,2);getline;print;getline;getline}}' >{}".format(
                                extraction_threads, f, out_here
                            ))
                            os.remove(f)
                            output_files.append(out_here)
                        elif 'fasta.gz' in output_format_possibilities:
                            logging.info("Converting {} to FASTA and compressing with pigz ..".format(f))
                            out_here = f.replace('.fastq.gz','.fasta.gz')
                            extern.run("pigz -cd {} |awk '{{print \">\" substr($0,2);getline;print;getline;getline}}' |pigz -p {} >{}".format(
                                f, extraction_threads, out_here
                            ))
                            os.remove(f)
                            output_files.append(out_here)
                        elif 'fastq' in output_format_possibilities:
                            logging.info("Decompressing {} with pigz ..".format(f))
                            extern.run("pigz -p {} -d {}".format(extraction_threads, f))
                            output_files.append(f.replace('.fastq.gz','.fastq'))
                        else:
                            raise Exception("Programming error")
                
    if not stdout and len(output_files) == 0:
        raise Exception("No output files found, something went amiss, unsure what.")

    logging.info("Output files: {}".format(', '.join(output_files)))

    
def extract(**kwargs):
    sra_file = kwargs.pop('sra_file')
    output_format_possibilities = kwargs.pop('output_format_possibilities',
        DEFAULT_OUTPUT_FORMAT_POSSIBILITIES)
    force = kwargs.pop('force', False)
    unsorted = kwargs.pop('unsorted', False)
    stdout = kwargs.pop('stdout', False)
    threads = kwargs.pop('threads',DEFAULT_THREADS)
    output_directory = kwargs.pop('output_directory', '.')

    if len(kwargs) > 0:
        raise Exception("Unexpected arguments detected: %s" % kwargs)

    if stdout and not unsorted:
        raise Exception("Currently --stdout must be used with --unsorted")

    run_identifier = os.path.basename(sra_file)
    if sra_file.endswith(".sra"):
        run_identifier = run_identifier[:-4]
    logging.debug("Using run identifier {}".format(run_identifier))

    output_location_factory = OutputLocation(output_directory)

    # Checking for already existing files
    if stdout:
        skip_download_and_extraction, output_files = False, []
    else:
        skip_download_and_extraction, output_files = _check_for_existing_files(
            output_location_factory, run_identifier, output_format_possibilities, force
        )
    
    if unsorted and stdout:
        format = output_format_possibilities[0]
        if format == 'fasta':
            logging.info("Extracting unsorted .sra file to STDOUT in FASTA format ..")
            cmd = "sracat {}".format(os.path.abspath(sra_file))
        elif format == 'fasta.gz':
            logging.info("Extracting unsorted .sra file to STDOUT in FASTA.GZ format ..")
            cmd = "sracat {} |pigz -p {} -c".format(os.path.abspath(sra_file), threads)
        elif format == 'fastq':
            logging.info("Extracting unsorted .sra file to STDOUT in FASTQ format ..")
            cmd = "sracat --qual {}".format(os.path.abspath(sra_file))
        elif format == 'fastq.gz':
            logging.info("Extracting unsorted .sra file to STDOUT in FASTQ.GZ format ..")
            cmd = "sracat --qual {} |pigz -p {} -c".format(os.path.abspath(sra_file), threads)
        else:
            raise Exception("Cannot extract with --stdout --unsorted format {}".format(format))
        logging.debug("Running command {}".format(cmd))
        try:
            subprocess.check_call(cmd, shell=True, stderr=subprocess.PIPE)
        except subprocess.CalledProcessError as e:
            raise Exception("Extraction of .sra to fasta format failed. Command run was '{}'. STDERR was '{}'".format(
                cmd, e.stderr
            ))
            
    elif unsorted and not stdout:
        def run_command(cmd):
            logging.debug("Running command {}".format(cmd))
            try:
                subprocess.check_call(cmd, shell=True, stderr=subprocess.PIPE)
            except subprocess.CalledProcessError as e:
                raise Exception(f"Extraction of .sra to format unsorted {format} failed. Command run was '{cmd}'. STDERR was '{e.stderr}'")

        # By default, we want separate outputs for forward and reverse.
        format = output_format_possibilities[0]
        if format == 'fasta':
            logging.info("Extracting .sra file to file(s) in unsorted FASTA format ..")
            cmd = f"sracat -o {output_location_factory.output_stem(run_identifier)} {os.path.abspath(sra_file)}"
            run_command(cmd)
            for name in ['x_1.fna','x_2.fna','x.fna']:
                f = output_location_factory.output_stem(name.replace('x',run_identifier))
                if os.path.exists(f):
                    new_name = re.sub(r'.fna$','.fasta', f)
                    os.rename(f, new_name)
                    output_files.append(new_name)
        elif format == 'fasta.gz':
            logging.info("Extracting .sra file to file(s) in unsorted FASTA.GZ format ..")
            cmd = f"sracat -o {run_identifier} {os.path.abspath(sra_file)}"

            # Make FIFOs so that we can use pigz instead of the slower built-in sracat -z.
            logging.debug("Creating FIFOs ..")
            os.mkfifo(f'{run_identifier}_1.fna')
            os.mkfifo(f'{run_identifier}_2.fna')
            os.mkfifo(f'{run_identifier}.fna')
            # Spawn pigz to read FIFOs.
            def spawn_pigz_it(in_name, out_name, threads):
                return subprocess.Popen(['bash','-c',f'pigz -c -p {threads} {in_name} > {out_name}'])
            pigz_commands = []
            for name in ['x_1.fna','x_2.fna','x.fna']:
                f = name.replace('x',run_identifier)
                output = output_location_factory.output_stem(re.sub('.fna$','.fasta.gz', f))
                pigz_commands.append([f, output, spawn_pigz_it(f'{f}', f"{output}", threads)])
            run_command(cmd)
            for (fifo, output, c) in pigz_commands:
                logging.debug(f"Waiting for pigz command {c.args} ..")

                # If sracat doesn't write to a fifo, then the pigz reading it
                # never finishes. So run another echo on top so at least 1 EOF
                # arrives to terminate the pigz process.
                logging.debug("Running echo to make sure at least one EOF arrived on the pipe")
                subprocess.Popen(['bash','-c',f'cat {fifo} > /dev/null'])
                extern.run(f'echo -n >> {fifo}')
                
                ret = c.wait()
                if ret != 0:
                    raise subprocess.SubprocessError(f"Command {c.args} returned with non-zero exitstatus {ret}")
                logging.debug("Process finished")
                os.remove(fifo)

                # Open the gzip file. If there is anything inside, keep it,
                # otherwise remove it as sracat never read it.
                remove_it = False
                with gzip.open(output) as f:
                    some = f.read(10)
                    if len(some) == 0:
                        logging.debug(f"Compressed file {output} is empty, removing")
                        remove_it = True
                if remove_it:
                    os.remove(output)

            for name in ['x_1.fasta.gz','x_2.fasta.gz','x.fasta.gz']:
                f = name.replace('x',run_identifier)
                if os.path.exists(f):
                    output_files.append(f)
        elif format == 'fastq':
            logging.info("Extracting .sra file to file(s) in unsorted FASTQ format ..")
            output_stem = output_location_factory.output_stem(run_identifier)
            cmd = f"sracat --qual -o {output_stem} {os.path.abspath(sra_file)}"
            run_command(cmd)
            for name in ['x_1.fastq','x_2.fastq','x.fastq']:
                f = name.replace('x',run_identifier)
                if os.path.exists(f):
                    output_files.append(f)
        elif format == 'fastq.gz':
            logging.info("Extracting .sra file to file(s) in unsorted FASTQ.GZ format ..")
            output_stem = output_location_factory.output_stem(run_identifier)
            cmd = f"sracat -z --qual -o {output_stem} {os.path.abspath(sra_file)}"
            run_command(cmd)
            for name in ['x_1.fastq.gz','x_2.fastq.gz','x.fastq.gz']:
                f = name.replace('x',run_identifier)
                if os.path.exists(f):
                    output_files.append(f)
        else:
            raise Exception("Cannot extract with --unsorted format {}".format(format))

    else:
        if not skip_download_and_extraction:
            logging.info("Extracting .sra file with fasterq-dump ..")

            # Change directory to the output directory using a "with", so that fasterq-dump outputs there, not here.
            sra_file_abs = os.path.abspath(sra_file)
            with bird_tool_utils.in_working_directory(output_directory):
                extern.run("fasterq-dump --threads {} {}".format(threads, sra_file_abs))

                if 'fastq' not in output_format_possibilities:
                    for fq in ['x_1.fastq','x_2.fastq','x.fastq']:
                        f = output_location_factory.output_stem(fq.replace('x',run_identifier))
                        if os.path.exists(f):
                            # Do the least work, currently we have FASTQ.
                            if 'fasta' in output_format_possibilities:
                                logging.info("Converting {} to FASTA ..".format(f))
                                out_here = output_location_factory.output_stem(re.sub('.fastq$','.fasta',f))
                                extern.run("awk '{{print \">\" substr($0,2);getline;print;getline;getline}}' {} >{}".format(
                                    f, out_here
                                ))
                                os.remove(f)
                                output_files.append(out_here)
                            elif 'fasta.gz' in output_format_possibilities:
                                logging.info("Converting {} to FASTA and compressing with pigz ..".format(f))
                                out_here = output_location_factory.output_stem(re.sub('.fastq$','.fasta.gz',f))
                                extern.run("awk '{{print \">\" substr($0,2);getline;print;getline;getline}}' {} |pigz -p {} >{}".format(
                                    f, threads, out_here
                                ))
                                os.remove(f)
                                output_files.append(out_here)
                            elif 'fastq.gz' in output_format_possibilities:
                                out_here = os.path.abspath(output_location_factory.output_stem(f'{f}.gz'))
                                logging.info("Compressing {} with pigz into {} ..".format(f, out_here))
                                extern.run("pigz -c -p {} {} > {}".format(threads, f, out_here))
                                os.remove(f)
                                output_files.append(out_here)
                            else:
                                raise Exception("Programming error")
                else:
                    for fq in ['x_1.fastq','x_2.fastq','x.fastq']:
                        f = fq.replace('x',run_identifier)
                        if os.path.exists(f):
                            output_files.append(f)

    return output_files

def gzip_test_files(gzip_files):
    """
    Run "pigz -t" on each result file, to check that it is a valid gzip file.

    Assumes the input is a list of paths of gzip files
    """
    for f in gzip_files:
        logging.info("Verifying gzip file {} ..".format(f))
        extern.run("pigz -t '{}'".format(f))


def annotate(**kwargs):
    run_identifiers = kwargs.pop('run_identifiers')
    run_identifiers_file = kwargs.pop('run_identifiers_file')
    bioproject_accessions = kwargs.pop('bioproject_accessions', None)  # kept for API stability
    bioproject_accession = kwargs.pop('bioproject_accession', None)
    output_file = kwargs.pop('output_file')
    output_format = kwargs.pop('output_format')
    all_columns = kwargs.pop('all_columns')

    if bioproject_accession and bioproject_accessions is None:
        bioproject_accessions = [bioproject_accession]

    num_inputs = 0
    if run_identifiers is not None: num_inputs += 1
    if run_identifiers_file is not None: num_inputs += 1
    if bioproject_accessions is not None: num_inputs += 1
    if num_inputs != 1:
        raise Exception("Must specify exactly one input type: --run-identifiers, --bioproject-accessions or --run-identifiers-list")
    
    if bioproject_accessions is not None:
        run_identifiers = SraMetadata().fetch_runs_from_bioprojects(bioproject_accessions)
        logging.debug("Found {} run(s) to annotate".format(len(run_identifiers)))
    if run_identifiers_file is not None:
        with open(run_identifiers_file) as f:
            run_identifiers = list([r.strip() for r in f.readlines()])
        
    if len(kwargs) > 0:
        raise Exception("Unexpected arguments detected: %s" % kwargs)

    metadata = SraMetadata().efetch_sra_from_accessions(run_identifiers)
    if metadata is None:
        logging.error("No runs to annotate")
        sys.exit(1)
    _output_formatted_metadata(metadata, output_file, output_format, all_columns)


def _output_formatted_metadata(metadata, output_file, output_format, all_columns):
    # NOTE: If changing this default set, also need to change the default set in human readable output below too.
    default_columns = [RUN_ACCESSION_KEY,BIOPROJECT_ACCESSION_KEY,'Gbp','library_strategy','library_selection','model',SAMPLE_NAME_KEY,'taxon_name']

    def prepare_for_tsv_csv(metadata, default_columns, all_columns):
        metadata_sorted = metadata.sort_values(RUN_ACCESSION_KEY)
        # For very large data frames, pandas throws an error 'InvalidIndexError:
        # Reindexing only valid with uniquely valued Index objects' when doing
        # the pd.concat() below. We have to do that concat because a simple
        # metadata_sorted['Gbp'] = ... gives a Performance warning. To get
        # around this, we reset the index to a RangeIndex, which does not
        # contain duplicates.
        metadata_sorted.reset_index(drop=True, inplace=True)
        metadata_sorted = pd.concat(
            [
                metadata_sorted,
                pd.DataFrame({'Gbp': [
                    round(bases/1e9, 3) if bases is not None else None for bases in metadata_sorted[BASES_KEY]]})
            ],
            axis=1)
        if all_columns:
            # Re-order columns to be consistent with human format output
            column_order = default_columns + [c for c in metadata_sorted.columns if c not in default_columns]
            return metadata_sorted[column_order]
        else:
            metadata_sorted = metadata_sorted[default_columns]
            return metadata_sorted

    output_path = sys.stdout if output_file is None else output_file

    if output_format == 'human':
        to_print = []
        for value in metadata[RUN_ACCESSION_KEY]:
            to_print.append({RUN_ACCESSION_KEY: value})
        for i, value in enumerate(metadata[BIOPROJECT_ACCESSION_KEY]):
            to_print[i][BIOPROJECT_ACCESSION_KEY] = value
        for i, value in enumerate(metadata[BASES_KEY]):
            to_print[i]['Gbp'] = "%.3f" % (value/1e9) if value is not None else None
        # for column in ['LibraryStrategy','LibrarySelection','Model','SampleName','ScientificName']:
        for column in ['library_strategy','library_selection','model',SAMPLE_NAME_KEY,'taxon_name']:
            for i, value in enumerate(metadata[column]):
                to_print[i][column] = value
        if all_columns:
            for col in metadata.columns:
                if col not in default_columns:
                    for i, value in enumerate(metadata[col]):
                        to_print[i][col] = value
        to_print = sorted(to_print, key=lambda x: x[RUN_ACCESSION_KEY])
        if output_path == sys.stdout:
            _printTable(sys.stdout, to_print)
        else:
            with open(output_path, 'w') as f:
                _printTable(f, to_print)
    elif output_format == 'csv':
        metadata_sorted = prepare_for_tsv_csv(metadata, default_columns, all_columns)
        metadata_sorted.to_csv(output_path, index=False)
    elif output_format == 'tsv':
        metadata_sorted = prepare_for_tsv_csv(metadata, default_columns, all_columns)
        metadata_sorted.to_csv(output_path, sep='\t', index=False)
    elif output_format == 'json':
        metadata_sorted = prepare_for_tsv_csv(metadata, default_columns, all_columns)
        metadata_sorted.to_json(output_path, orient='records', indent=2)
    elif output_format == 'feather':
        metadata_sorted = prepare_for_tsv_csv(metadata, default_columns, all_columns)
        with open(output_file,'wb') as f:
            metadata_sorted.to_feather(f)
    elif output_format == 'parquet':
        metadata_sorted = prepare_for_tsv_csv(metadata, default_columns, all_columns)
        with open(output_file,'wb') as f:
            metadata_sorted.to_parquet(f, index=False)
    else:
        raise Exception("Unexpected output format: {}".format(output_format))

def _printTable(output_stream, myDict, colList=None):
   if not colList: colList = list(myDict[0].keys() if myDict else [])
   myList = [colList] # 1st row = header
   for item in myDict: myList.append([str(item[col] if item[col] is not None else '') for col in colList])
   colSize = [max(map(len,col)) for col in zip(*myList)]
   formatStr = ' | '.join(["{{:<{}}}".format(i) for i in colSize])
   myList.insert(1, ['-' * i for i in colSize]) # Seperating line
   for item in myList: print(formatStr.format(*item), file=output_stream)

def _check_for_existing_files(output_location_factory, run_identifier, output_format_possibilities, force):
    skip_download_and_extraction = False
    output_files = []

    def maybe_skip_or_force(path, output_files, force):
        skip_download_and_extraction = False
        final_path = output_location_factory.output_stem(path)
        if os.path.exists(final_path):
            if force:
                logging.warn("Removing previous file {}".format(final_path))
                os.remove(final_path)
            else:
                skip_download_and_extraction = True
                output_files.append(final_path)
                logging.info(
                    "Skipping download/extraction of {} as an output file already appears to exist, as file {}".format(run_identifier, final_path))
        return skip_download_and_extraction, output_files

    for file_type in output_format_possibilities:
        if file_type == 'sra':
            path = "{}.{}".format(run_identifier, file_type)
            skip, output_files = maybe_skip_or_force(path, output_files, force)
            if skip: skip_download_and_extraction = True
        elif file_type == 'fastq':
            possibilities = ['x.fastq','x_1.fastq','x_2.fastq']
            for path in possibilities:
                skip, output_files = maybe_skip_or_force(path.replace('x',run_identifier), output_files, force)
                if skip: skip_download_and_extraction = True
        elif file_type == 'fastq.gz':
            possibilities = ['x.fastq.gz','x_1.fastq.gz','x_2.fastq.gz']
            for path in possibilities:
                skip, output_files = maybe_skip_or_force(path.replace('x',run_identifier), output_files, force)
                if skip: skip_download_and_extraction = True
        elif file_type == 'fasta':
            possibilities = ['x.fasta','x_1.fasta','x_2.fasta']
            for path in possibilities:
                skip, output_files = maybe_skip_or_force(path.replace('x',run_identifier), output_files, force)
                if skip: skip_download_and_extraction = True
        elif file_type == 'fasta.gz':
            possibilities = ['x.fasta.gz','x_1.fasta.gz','x_2.fasta.gz']
            for path in possibilities:
                skip, output_files = maybe_skip_or_force(path.replace('x',run_identifier), output_files, force)
                if skip: skip_download_and_extraction = True
        else:
            raise Exception("Programming error")

    return skip_download_and_extraction, output_files


import logging
import hashlib

class MD5:
    @staticmethod
    def check_md5sum(file_path, expected_md5sum):
        logging.debug("Checking md5sum for {} ..".format(file_path))
        
        hash_md5 = hashlib.md5()
        with open(file_path, "rb") as f:
            for chunk in iter(lambda: f.read(4096), b""):
                hash_md5.update(chunk)
        return hash_md5.hexdigest()

        if hash_md5 == expected_md5sum:
            return True
        else:
            return False


#!/usr/bin/env python3

#=======================================================================
# Authors: Ben Woodcroft
#
# Unit tests.
#
# Copyright
#
# This is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License.
# If not, see <http://www.gnu.org/licenses/>.
#=======================================================================

import unittest
import os.path
import sys

import extern

sys.path = [os.path.join(os.path.dirname(os.path.realpath(__file__)),'..')]+sys.path
kingfisher = os.path.join(os.path.dirname(os.path.realpath(__file__)),'..','bin','kingfisher')

from bird_tool_utils import in_tempdir

class Tests(unittest.TestCase):
    maxDiff = None
    
    def test_fastq_gz(self):
        cmd_stub = '{} get -r SRR12118866 -m'.format(kingfisher)
        with in_tempdir():
            extern.run("{} {}".format(cmd_stub,'ena-ascp'))
            self.assertTrue(os.path.getsize('SRR12118866_1.fastq.gz')==4117481)
            self.assertTrue(os.path.getsize('SRR12118866_2.fastq.gz')==4945891)

    def test_fastq_via_ena_ascp(self):
        with in_tempdir():
            extern.run('{} get -r SRR12118866 -m ena-ascp --output-format-possibilities fastq'.format(
                kingfisher))
            self.assertTrue(os.path.getsize('SRR12118866_1.fastq')==19930812)
            self.assertTrue(os.path.getsize('SRR12118866_2.fastq')==19930812)

    def test_fasta_via_ena_ascp(self):
        with in_tempdir():
            extern.run('{} get -r SRR12118866 -m ena-ascp --output-format-possibilities fasta.gz fasta'.format(
                kingfisher))
            self.assertTrue(os.path.getsize('SRR12118866_1.fasta')==10391100)
            self.assertTrue(os.path.getsize('SRR12118866_2.fasta')==10391100)

    def test_fasta_gz_via_ena_ascp(self):
        with in_tempdir():
            extern.run('{} get -r SRR12118866 -m ena-ascp --output-format-possibilities fasta.gz'.format(
                kingfisher))
            self.assertTrue(os.path.getsize('SRR12118866_1.fasta.gz')==746749)
            self.assertTrue(os.path.getsize('SRR12118866_2.fasta.gz')==899862)

    def test_fasta_gz_via_ena_ftp(self):
        '''not sure why but this fails on travis'''
        with in_tempdir():
            extern.run('{} get -r SRR12118866 -m ena-ftp --output-format-possibilities fasta.gz'.format(
                kingfisher))
            self.assertTrue(os.path.getsize('SRR12118866_1.fasta.gz')==746749)
            self.assertTrue(os.path.getsize('SRR12118866_2.fasta.gz')==899862)

    def test_fasta_gz_via_ena_ftp_output_directory(self):
        with in_tempdir():
            extern.run('{} get -r SRR12118866 -m ena-ftp --output-format-possibilities fasta.gz --output-directory outdir'.format(
                kingfisher))
            self.assertTrue(os.path.getsize('outdir/SRR12118866_1.fasta.gz')==746749)
            self.assertTrue(os.path.getsize('outdir/SRR12118866_2.fasta.gz')==899862)

    def test_aria2_aws_fasta(self):
        '''cannot test this in travis because conda aria2 is broken'''
        with in_tempdir():
            extern.run('{} get --download-threads 4 -r SRR12118866 -m aws-http --output-format-possibilities fasta.gz'.format(
                kingfisher))
            self.assertTrue(os.path.getsize('SRR12118866_1.fasta.gz')==757641)
            self.assertTrue(os.path.getsize('SRR12118866_2.fasta.gz')==907591)

    def test_aria2_aws_fasta_output_directory(self):
        '''cannot test this in travis because conda aria2 is broken'''
        with in_tempdir():
            extern.run('{} get --download-threads 4 -r SRR12118866 -m aws-http --output-format-possibilities fasta.gz --output-directory outdir'.format(
                kingfisher))
            self.assertTrue(os.path.getsize('outdir/SRR12118866_1.fasta.gz')==757641)
            self.assertTrue(os.path.getsize('outdir/SRR12118866_2.fasta.gz')==907591)

    def test_aria2_ena_fastq_gz(self):
        '''cannot test this in travis because conda aria2 is broken'''
        with in_tempdir():
            extern.run('{} get --download-threads 4 -r SRR12118866 -m ena-ftp -f fastq.gz'.format(
                kingfisher))
            self.assertTrue(os.path.getsize('SRR12118866_1.fastq.gz')==4117481)
            self.assertTrue(os.path.getsize('SRR12118866_2.fastq.gz')==4945891)

    def test_aria2_ena_fastq_gz_check_md5sum(self):
        '''cannot test this in travis because conda aria2 is broken'''
        with in_tempdir():
            extern.run('{} get --download-threads 4 -r SRR12118866 -m ena-ftp -f fastq.gz --check-md5sums 2>kingfisher_stderr'.format(
                kingfisher))
            self.assertTrue(os.path.getsize('SRR12118866_1.fastq.gz')==4117481)
            self.assertTrue(os.path.getsize('SRR12118866_2.fastq.gz')==4945891)
            with open('kingfisher_stderr') as f:
                stderr = f.read()
                self.assertTrue('MD5sum OK for SRR12118866_1.fastq.gz' in stderr)
                self.assertTrue('MD5sum OK for SRR12118866_2.fastq.gz' in stderr)
        

if __name__ == "__main__":
    unittest.main()



from io import StringIO
import subprocess
import logging
import os
import pandas as pd

import extern
import bird_tool_utils

from .md5sum import MD5

DEFAULT_LINUX_ASPERA_SSH_KEY_LOCATION = os.path.join(os.path.dirname(os.path.realpath(__file__)),'data','asperaweb_id_dsa.openssh')

class EnaFileReport:
    def __init__(self, file_paths, md5sums):
        self.file_paths = file_paths
        self.md5sums = md5sums

class EnaDownloader:
    def get_ftp_download_urls(self, run_id):
        # Get the textual representation of the run. We specifically need the
        # fastq_ftp bit, and the MD5
        logging.info("Querying ENA for FTP paths for {}..".format(run_id))
        query_url = "https://www.ebi.ac.uk/ena/portal/api/filereport?accession={}&" \
            "result=read_run&fields=fastq_ftp,fastq_md5".format(
            run_id)
        logging.debug("Querying '{}'".format(query_url))
        text = extern.run("curl --silent '{}'".format(query_url))

        header = True
        logging.debug("Found text from ENA API: {}".format(text))

        df = pd.read_csv(StringIO(text), sep='\t', header=0, index_col=False)

        # Expect just 1 row
        if len(df) == 0:
            # One (current) example of this is DRR086621
            logging.error(
                "No FTP download URLs found for run {}, cannot continue".format(
                    run_id))
            return False
        elif len(df) != 1:
            logging.error("Expected 1 row from ENA API for accession {}, got {}".format(run_id, len(df)))
            return False

        for _, row in df.iterrows():
            # e.g. ERR1346134 at time of writing. See https://github.com/wwood/kingfisher-download/issues/25
            if isinstance(float("nan"), type(row['fastq_ftp'])):
                logging.error("No ENA FTP download URLs found for run {}, cannot continue".format(run_id))
                return False
            ftp_urls = row['fastq_ftp'].split(';')
            md5sums = row['fastq_md5'].split(';')
            logging.debug("Found {} FTP URLs for download: {}".format(
                len(ftp_urls), ", ".join(ftp_urls)))

        return EnaFileReport(ftp_urls, md5sums)

    def _clean_incomplete_files(self, paths):
        for path in paths:
            if os.path.exists(path):
                logging.info("Removing file that is either incomplete or part of an incomplete pair: {}".format(path))
                os.remove(path)

    def download_with_aspera(self, run_id, output_directory, quiet=False, ascp_args='', ssh_key=None, check_md5sums=False):
        if ssh_key is None:
            logging.debug("Attempting to find aspera ssh key file at {}".format(DEFAULT_LINUX_ASPERA_SSH_KEY_LOCATION))
            if os.path.exists(DEFAULT_LINUX_ASPERA_SSH_KEY_LOCATION):
                ssh_key_file = DEFAULT_LINUX_ASPERA_SSH_KEY_LOCATION
            else:
                raise Exception("Cannot find aspera ssh key file, please specify with --aspera-ssh-key")
        else:
            ssh_key_file = ssh_key
        logging.info("Using aspera ssh key file: {}".format(ssh_key_file))

        report = self.get_ftp_download_urls(run_id)
        if report is False:
            return False
        ftp_urls = report.file_paths
        md5sums = report.md5sums

        logging.info("Downloading {} FTP read set(s): {}".format(
            len(ftp_urls), ", ".join(ftp_urls)))

        output_files = []
        for url, md5 in zip(ftp_urls, md5sums):
            quiet_args = ''
            if quiet:
                quiet_args = ' -Q'
            output_file = os.path.join(output_directory, os.path.basename(url))
            logging.debug("Getting output file {}".format(output_file))
            cmd = "ascp{} -T -l 300m -P33001 {} -i {} era-fasp@fasp.sra.ebi.ac.uk:{} {}".format(
                quiet_args,
                ascp_args,
                ssh_key_file,
                url.replace('ftp.sra.ebi.ac.uk', ''),
                output_directory)
            logging.info("Running command: {}".format(cmd))
            try:
                extern.run(cmd)
            except Exception as e:
                logging.warn("Error downloading from ENA with ASCP: {}".format(e))
                self._clean_incomplete_files(output_files+[output_file])
                return False
            if check_md5sums:
                if MD5.check_md5sum(output_file, md5):
                    logging.info("MD5sum OK for {}".format(output_file))
                else:
                    logging.error("MD5sum failed for {}".format(output_file))
                    self._clean_incomplete_files(output_files+[output_file])
                    return False
            output_files.append(output_file)
        return output_files

    def download_with_curl(self, run_id, num_threads, output_directory, check_md5sums=False):
        with bird_tool_utils.in_working_directory(output_directory):
            report = self.get_ftp_download_urls(run_id)
            if report is False:
                return False
            ftp_urls = report.file_paths
            md5sums = report.md5sums

            downloaded = []
            for url, md5 in zip(ftp_urls, md5sums):
                logging.info("Downloading {} ..".format(url))
                output_file = os.path.basename(url)
                if num_threads > 1:
                    cmd = "aria2c -x{} -o {} 'ftp://{}'".format(
                        num_threads, output_file, url)
                else:
                    cmd = "curl -L '{}' -o {}".format(url, output_file)
                try:
                    subprocess.check_call(cmd, shell=True)
                except subprocess.CalledProcessError as e:
                    logging.warning("Method ena-ftp failed, error was {}".format(e))
                    self._clean_incomplete_files(downloaded+[output_file])
                    return False
                
                if check_md5sums:
                    if MD5.check_md5sum(output_file, md5):
                        logging.info("MD5sum OK for {}".format(output_file))
                    else:
                        logging.error("MD5sum failed for {}".format(output_file))
                        self._clean_incomplete_files(downloaded+[output_file])
                        return False
                downloaded.append(os.path.join(output_directory, output_file))
        return downloaded




import logging
import json
import re

import extern

from .exception import DownloadMethodFailed


class Location:
    @staticmethod
    def get_ncbi_locations(run_id):
        json_location_string = 'https://locate.ncbi.nlm.nih.gov/sdl/2/retrieve?&acc={}&accept-alternate-locations=yes'.format(
            run_id)
        json_response = extern.run('curl -q \'{}\''.format(json_location_string))
        logging.debug("Got location JSON: {}".format(json_response))

        j = json.loads(json_response)
        if 'version' not in j or j['version'] != '2':
            raise Exception(
                "Unexpected json location string returned: {}", json_location_string)
        # TODO: Assumes there is only 1 result, which is all I've ever seen
        return NcbiLocationJson(j)


class AwsLocation:
    def __init__(self, object_json, location_json):
        self.object_json = object_json
        self.j = location_json

    def service(self):
        if self.j['service'] == 's3':
            # The run ERR209516 for instance is non-ODP at s3://sra-pub-run-3/ERR209516/ERR209516.2
            # whereas SRR12324253 is Open Data, at https://sra-pub-run-odp.s3.amazonaws.com/sra/SRR12324253/SRR12324253
            # SRR12316190 is https://sra-pub-sars-cov2.s3.amazonaws.com/run/SRR12316190/SRR12316190
            if 'sra-pub-run-odp' in self.j['link']:
                return 's3-odp'
            elif 'sra-pub-sars-cov2' in self.j['link']:
                return 's3-sars-cov2'
            else:
                return 's3-pay'
        else:
            raise Exception("Programming error: {}".format(self.j))

    def s3_command_prefix(self, run_id):
        if self.service() == 's3-pay':
            # bucket/key not present in paid link from
            # 'https://locate.ncbi.nlm.nih.gov/sdl/2/retrieve?&acc=ERR209516&accept-alternate-locations=yes'
            # for instance.
            if 'bucket' in self.j and 'key' in self.j:
                # Below return is not tested currently, because I do not currently
                # know of an SRA accession that falls into this category.
                return 'aws s3 cp s3://{}/{}'.format(self.j['bucket'], self.j['key'])
            else:
                raise DownloadMethodFailed("Unexpected form of S3 location JSON: {}".format(self.j))
                # elif 'link' in self.j:
                # E.g. SRR16940109 is currently:
                # {
                #     "service": "s3",
                #     "region": "us-east-1",
                #     "payRequired": true,
                #     "link": "https://sra-pub-run-6.s3.amazonaws.com/SRR16940109/SRR16940109.1"
                # }
                # => That link is currently not available
            
        elif self.service() == 's3-odp':
            # Use --no-sign-request to avoid the AWS CLI signing into an
            # account, avoiding potential usage charges. There is a possibility
            # here that a non-ODP link is specified in the location API, but
            # this will only case an error since we are using --no-sign-request.
            return 'aws s3 cp --no-sign-request s3://sra-pub-run-odp/sra/{}/{}'.format(run_id, run_id)
        elif self.service() == 's3-sars-cov2':
            return 'aws s3 cp --no-sign-request s3://sra-pub-sars-cov2/run/{}/{}'.format(run_id, run_id)
        else:
            raise Exception("Unexpected json location found: {}", self.j)

    def link(self):
        return self.j['link']

    def md5sum(self):
        return self.object_json['md5']


class GcpLocation:
    def __init__(self, object_json, location_json):
        self.object_json = object_json
        self.j = location_json

    def gs_path(self):
        if 'rehydrationRequired' in self.j and self.j['rehydrationRequired'] == True:
            raise DownloadMethodFailed("Rehydration required from GCP, so ignoring")
        elif 'key' in self.j and 'bucket' in self.j:
            # Possibly now outdated and this branch is never used?
            return 'gs://{}/{}'.format(self.j['bucket'], self.j['key'])
        elif 'link' in self.j:
            r = re.compile('https://storage.googleapis.com/(.*?)/(.*)')
            m = r.match(self.j['link'])
            if m is None:
                raise DownloadMethodFailed("Unexpected GCP link URL in {}".format(self.j))
            else:
                return 'gs://{}/{}'.format(m[1],m[2])
        else:
            raise DownloadMethodFailed("Unsure how to copy from GCP location {}".format(self.j))

    def md5sum(self):
        return self.object_json['md5']


class NcbiLocationJson:
    OBJECT_TYPE_SRA = 'sra-qual'
    OBJECT_TYPE_SRA_NOQUAL = 'sra-noqual'

    GCP_SERVICE = 'gs-service'
    AWS_SERVICE = 's3-service'

    def __init__(self, j):
        self.j = j

    def object_locations(self, object_type, service, allow_paid):
        # First get a set of objects that are suitable
        passable_objects = []
        if 'files' not in self.j['result'][0]:
            error_msg = "No results returned from NCBI location API"
            if 'msg' in self.j['result'][0]:
                error_msg += ". msg was '{}'".format(self.j['result'][0]['msg'])
            logging.warning(error_msg)
            return []
        for obj in self.j['result'][0]['files']:
            if obj['type'] == 'sra':
                if obj['name'].endswith('.noqual'):
                    if object_type == NcbiLocationJson.OBJECT_TYPE_SRA_NOQUAL:
                        passable_objects.append(obj)
                else:
                    if object_type == NcbiLocationJson.OBJECT_TYPE_SRA:
                        passable_objects.append(obj)
        # Then get a set of locations of those objects that suit
        passable_objects_and_locations = []
        for obj in passable_objects:
            for loc in obj['locations']:
                logging.debug("Assessing location {}".format(loc))
                if 'payRequired' in loc and loc['payRequired'] != False and allow_paid != True:
                    # Payment is required but we aren't allow-paid
                    logging.debug("Excluding location as we aren't paid: {}".format(loc))
                    continue
                elif service == NcbiLocationJson.GCP_SERVICE:
                    if loc['service'] == 'gs':
                        logging.debug("Accepting location")
                        passable_objects_and_locations.append(GcpLocation(obj, loc))
                    else:
                        logging.debug("Location has the wrong service")
                elif service == NcbiLocationJson.AWS_SERVICE:
                    if loc['service'] == 's3':
                        logging.debug("Accepting location")
                        passable_objects_and_locations.append(AwsLocation(obj, loc))
                    else:
                        logging.debug("Location has the wrong service")
                else:
                    logging.debug("Discarding location as unsuitable: {}".format(loc))
        return passable_objects_and_locations


import io
from os.path import dirname, join
from setuptools import setup


# read the contents of your README file
from os import path
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()


def get_version(relpath):
  """Read version info from a file without importing it"""
  for line in io.open(join(dirname(__file__), relpath), encoding="cp437"):
    if "__version__" in line:
      if '"' in line:
        return line.split('"')[1]
      elif "'" in line:
        return line.split("'")[1]


setup(
    name='kingfisher',
    version=get_version("kingfisher/version.py"),
    url='https://github.com/wwood/kingfisher-download',
    license='GPL3+',
    author='Ben Woodcroft',
    author_email='benjwoodcroft@gmail.com',
    description='Download/extract biological FASTA/Q read data and metadata',
    long_description=long_description,
    long_description_content_type='text/markdown',
    packages=['kingfisher'],
    package_data={'kingfisher': [
            'data/asperaweb_id_dsa.openssh',
                       ]},
    data_files=[(".", ["README.md", "LICENCE.txt"])],
    include_package_data=True,
    install_requires= [
      'extern',
      'requests',
      'tqdm',
      'pandas',
      'bird_tool_utils',
      'pyarrow',
    ],
    scripts=['bin/kingfisher'],
    classifiers=["Topic :: Scientific/Engineering :: Bio-Informatics"],
)



#!/usr/bin/env python3

#=======================================================================
# Authors: Ben Woodcroft
#
# Unit tests.
#
# Copyright
#
# This is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License.
# If not, see <http://www.gnu.org/licenses/>.
#=======================================================================

import unittest
import os.path
import sys
import subprocess
import pytest

import extern

sys.path = [os.path.join(os.path.dirname(os.path.realpath(__file__)),'..')]+sys.path
sys.path = [os.path.join(os.path.dirname(os.path.realpath(__file__)),'..','..')]+sys.path
kingfisher = os.path.join(os.path.dirname(os.path.realpath(__file__)),'..','bin','kingfisher')

from bird_tool_utils import in_tempdir

class Tests(unittest.TestCase):
    maxDiff = None
    
    def test_unpaid_methods(self):
        cmd_stub = '{} get -r SRR12118866 -m'.format(kingfisher)
        for method in ('aws-http','prefetch'):
            with in_tempdir():
                extern.run("{} {}".format(cmd_stub,method))
                self.assertTrue(os.path.getsize('SRR12118866_1.fastq')==21411192)
                self.assertTrue(os.path.getsize('SRR12118866_2.fastq')==21411192)

    def test_unpaid_methods_outdir(self):
        cmd_stub = '{} get -r SRR12118866 --output-directory outdir -m'.format(kingfisher)
        for method in ('aws-http','prefetch'):
            with in_tempdir():
                extern.run("{} {}".format(cmd_stub,method))
                self.assertTrue(os.path.getsize('outdir/SRR12118866_1.fastq')==21411192)
                self.assertTrue(os.path.getsize('outdir/SRR12118866_2.fastq')==21411192)

    def test_fasta_via_sra(self):
        with in_tempdir():
            extern.run('{} get -r SRR12118866 -m aws-http --output-format-possibilities fasta.gz fasta'.format(
                kingfisher))
            self.assertTrue(os.path.getsize('SRR12118866_1.fasta')==10705596)
            self.assertTrue(os.path.getsize('SRR12118866_2.fasta')==10705596)

    def test_fasta_via_sra_outdir(self):
        with in_tempdir():
            extern.run('{} get -r SRR12118866 -m aws-http --output-directory outdir --output-format-possibilities fasta.gz fasta'.format(
                kingfisher))
            self.assertTrue(os.path.getsize('outdir/SRR12118866_1.fasta')==10705596)
            self.assertTrue(os.path.getsize('outdir/SRR12118866_2.fasta')==10705596)

    def test_fasta_gz_via_sra(self):
        with in_tempdir():
            extern.run('{} get -r SRR12118866 -m aws-http --output-format-possibilities fasta.gz'.format(
                kingfisher))
            self.assertTrue(os.path.getsize('SRR12118866_1.fasta.gz')==757641)
            self.assertTrue(os.path.getsize('SRR12118866_2.fasta.gz')==907591)

    def test_fasta_gz_via_sra_outdir(self):
        with in_tempdir():
            extern.run('{} get -r SRR12118866 -m aws-http --output-directory outdir --output-format-possibilities fasta.gz'.format(
                kingfisher))
            self.assertTrue(os.path.getsize('outdir/SRR12118866_1.fasta.gz')==757641)
            self.assertTrue(os.path.getsize('outdir/SRR12118866_2.fasta.gz')==907591)

    def test_sra_via_aws(self):
        with in_tempdir():
            extern.run('{} get -r SRR12118866 -m aws-http --output-format-possibilities sra'.format(
                kingfisher))
            self.assertTrue(os.path.getsize('SRR12118866.sra')==11643188)

    def test_sra_via_aws_outdir(self):
        with in_tempdir():
            extern.run('{} get -r SRR12118866 -m aws-http --output-directory outdir --output-format-possibilities sra'.format(
                kingfisher))
            self.assertTrue(os.path.getsize('outdir/SRR12118866.sra')==11643188)

    def test_aws_guess_location(self):
        with in_tempdir():
            extern.run('{} get -r SRR12118866 -m aws-http --output-format-possibilities sra --guess_aws_location'.format(
                kingfisher))
            self.assertTrue(os.path.getsize('SRR12118866.sra')==11643188)

    @pytest.mark.aws_cp
    def test_aws_cp_covid(self):
        with in_tempdir():
            extern.run('{} get -r ERR6167542 --force -f sra -m aws-cp'.format(kingfisher))
            self.assertTrue(os.path.getsize('ERR6167542.sra')==3463862)

    def test_stdout_unsorted_fasta_via_sra(self):
        with in_tempdir():
            self.assertEqual('e53aeb5b0ae367d24bea4023ce940eea  -\n',
                extern.run('{} get -r SRR12118866 -m aws-http --output-format-possibilities fasta --stdout --unsorted |md5sum'.format(
                    kingfisher
                )))

    def test_extract_stdout_fasta(self):
        with in_tempdir():
            extern.run('{} get -r SRR12118866 -m aws-http --output-format-possibilities sra'.format(kingfisher))
            self.assertEqual('e53aeb5b0ae367d24bea4023ce940eea  -\n',
                extern.run('{} extract --sra SRR12118866.sra --output-format-possibilities fasta --stdout --unsorted |md5sum'.format(
                    kingfisher
                )))

    def test_unsorted_get_fasta_file_output(self):
        with in_tempdir():
            extern.run('{} get -r SRR12118866 -m aws-http --output-format-possibilities fasta --unsorted'.format(kingfisher))
            self.assertTrue(os.path.getsize('SRR12118866_1.fasta')==10122654)
            self.assertTrue(os.path.getsize('SRR12118866_2.fasta')==10122654)
            self.assertFalse(os.path.exists('SRR12118866.fasta'))

    def test_unsorted_extract_file_outputs_fasta_uncompressed(self):
        sra = os.path.abspath(f"test/data/SRR12118866.sra")
        
        with in_tempdir():
            extern.run('{} extract --sra {} --output-format-possibilities fasta --unsorted'.format(kingfisher, sra))
            self.assertTrue(os.path.getsize('SRR12118866_1.fasta')==10122654)
            self.assertTrue(os.path.getsize('SRR12118866_2.fasta')==10122654)
            self.assertFalse(os.path.exists('SRR12118866.fasta'))

    def test_unsorted_extract_file_outputs_fasta_uncompressed_output_directory(self):
        sra = os.path.abspath(f"test/data/SRR12118866.sra")

        with in_tempdir():
            extern.run('{} extract --sra {} --output-format-possibilities fasta --unsorted --output-directory outdir'.format(kingfisher, sra))
            self.assertTrue(os.path.getsize('outdir/SRR12118866_1.fasta')==10122654)
            self.assertTrue(os.path.getsize('outdir/SRR12118866_2.fasta')==10122654)
            self.assertFalse(os.path.exists('outdir/SRR12118866.fasta'))

    def test_unsorted_extract_file_outputs_fasta_gz(self):
        sra = f"test/data/SRR12118866.sra"

        if os.path.exists('SRR12118866_1.fasta.gz'):
            os.remove('SRR12118866_1.fasta.gz')
        if os.path.exists('SRR12118866_2.fasta.gz'):
            os.remove('SRR12118866_2.fasta.gz')

        cmd = '{} extract --sra {} --output-format-possibilities fasta.gz --unsorted'.format(kingfisher, sra)
        # For reasons I don't understand, running this via extern hangs when running this test specifically on b2
        subprocess.check_call(cmd.split(' ')) 
        self.assertEqual('fb284c28aac4513249b196ec75dc3c8d  -\n', extern.run('pigz -cd SRR12118866_1.fasta.gz |md5sum'))
        self.assertEqual('311f8898bd6d575ae3ec6a7188b08836  -\n', extern.run('pigz -cd SRR12118866_2.fasta.gz |md5sum'))
        self.assertFalse(os.path.exists('SRR12118866.fasta.gz'))
        os.remove('SRR12118866_1.fasta.gz')
        os.remove('SRR12118866_2.fasta.gz')

    def test_unsorted_extract_file_outputs_fasta_gz_output_directory(self):
        sra = os.path.abspath(f"test/data/SRR12118866.sra")

        with in_tempdir():
            cmd = '{} extract --sra {} --output-format-possibilities fasta.gz --unsorted --output-directory outdir'.format(kingfisher, sra)
            # For reasons I don't understand, running this via extern hangs when running this test specifically on b2
            subprocess.check_call(cmd.split(' ')) 
            self.assertEqual('fb284c28aac4513249b196ec75dc3c8d  -\n', extern.run('pigz -cd outdir/SRR12118866_1.fasta.gz |md5sum'))
            self.assertEqual('311f8898bd6d575ae3ec6a7188b08836  -\n', extern.run('pigz -cd outdir/SRR12118866_2.fasta.gz |md5sum'))
            self.assertFalse(os.path.exists('SRR12118866.fasta.gz'))

    def test_unsorted_extract_file_outputs_fastq_uncompressed(self):
        sra = os.path.abspath(f"test/data/SRR12118866.sra")

        with in_tempdir():
            extern.run('{} extract --sra {} --output-format-possibilities fastq --unsorted'.format(kingfisher, sra))
            self.assertTrue(os.path.getsize('SRR12118866_1.fastq')==19662366)
            self.assertTrue(os.path.getsize('SRR12118866_2.fastq')==19662366)
            self.assertFalse(os.path.exists('SRR12118866.fastq'))

    def test_unsorted_extract_file_outputs_fastq_uncompressed_output_directory(self):
        sra = os.path.abspath(f"test/data/SRR12118866.sra")

        with in_tempdir():
            extern.run('{} extract --sra {} --output-format-possibilities fastq --output-directory outdir --unsorted'.format(kingfisher, sra))
            self.assertTrue(os.path.getsize('outdir/SRR12118866_1.fastq')==19662366)
            self.assertTrue(os.path.getsize('outdir/SRR12118866_2.fastq')==19662366)
            self.assertFalse(os.path.exists('outdir/SRR12118866.fastq'))

    def test_unsorted_extract_file_outputs_fastq_gz(self):
        sra = os.path.abspath(f"test/data/SRR12118866.sra")

        with in_tempdir():
            extern.run('{} extract --sra {} --output-format-possibilities fastq.gz --unsorted'.format(kingfisher, sra))
            self.assertTrue(os.path.getsize('SRR12118866_1.fastq.gz')==4009949)
            self.assertTrue(os.path.getsize('SRR12118866_2.fastq.gz')==4834456)
            self.assertFalse(os.path.exists('SRR12118866.fastq.gz'))

    def test_unsorted_extract_file_outputs_fastq_gz_output_directory(self):
        sra = os.path.abspath(f"test/data/SRR12118866.sra")

        with in_tempdir():
            extern.run('{} extract --sra {} --output-format-possibilities fastq.gz --output-directory outdir --unsorted'.format(kingfisher, sra))
            self.assertTrue(os.path.getsize('outdir/SRR12118866_1.fastq.gz')==4009949)
            self.assertTrue(os.path.getsize('outdir/SRR12118866_2.fastq.gz')==4834456)
            self.assertFalse(os.path.exists('outdir/SRR12118866.fastq.gz'))

    def test_extract_fastq(self):
        with in_tempdir():
            extern.run('{} get -r SRR12118866 -m aws-http --output-format-possibilities sra'.format(kingfisher))
            extern.run('{} extract --sra SRR12118866.sra --output-format-possibilities fastq'.format(
                kingfisher
            ))
            self.assertTrue(os.path.getsize('SRR12118866_1.fastq')==21411192)
            self.assertTrue(os.path.getsize('SRR12118866_2.fastq')==21411192)

    def test_extract_fastq_output_directory(self):
        with in_tempdir():
            extern.run('{} get -r SRR12118866 -m aws-http --output-directory outdir --output-format-possibilities sra'.format(kingfisher))
            extern.run('{} extract --sra outdir/SRR12118866.sra --output-format-possibilities fastq --output-directory outdir'.format(
                kingfisher
            ))
            self.assertTrue(os.path.getsize('outdir/SRR12118866_1.fastq')==21411192)
            self.assertTrue(os.path.getsize('outdir/SRR12118866_2.fastq')==21411192)

    def test_extract_fastq_no_force(self):
        with in_tempdir():
            extern.run('{} get -r SRR12118866 -m aws-http --output-format-possibilities sra'.format(kingfisher))
            # extern.run('touch SRR12118866.sra')
            extern.run('touch SRR12118866.fastq SRR12118866_1.fastq SRR12118866_2.fastq'.format(kingfisher))
            r = subprocess.run(['bash','-c','{} extract --sra SRR12118866.sra --output-format-possibilities fastq'.format(
                kingfisher
                )],
                stderr=subprocess.PIPE,
                check=True)
            self.assertTrue('SRR12118866 as an output file already appears to exist' in r.stderr.decode())

    def test_extract_fastq_force(self):
        with in_tempdir():
            extern.run('{} get -r SRR12118866 -m aws-http --output-format-possibilities sra'.format(kingfisher))
            # extern.run('touch SRR12118866.sra')
            extern.run('touch SRR12118866.fastq SRR12118866_1.fastq SRR12118866_2.fastq'.format(kingfisher))
            r = subprocess.run(['bash','-c','{} extract --sra SRR12118866.sra --output-format-possibilities fastq --force'.format(
                kingfisher
                )],
                stderr=subprocess.PIPE,
                check=True)
            self.assertTrue(os.path.getsize('SRR12118866_1.fastq')==21411192)
            self.assertTrue(os.path.getsize('SRR12118866_2.fastq')==21411192)
            self.assertFalse('SRR12118866 as an output file already appears to exist' in r.stderr.decode())

    def test_extract_fastq_force_output_directory(self):
        with in_tempdir():
            extern.run('{} get -r SRR12118866 -m aws-http --output-format-possibilities sra --output-directory outdir'.format(kingfisher))
            # extern.run('touch SRR12118866.sra')
            extern.run('touch outdir/SRR12118866.fastq outdir/SRR12118866_1.fastq outdir/SRR12118866_2.fastq'.format(kingfisher))
            r = subprocess.run(['bash','-c','{} extract --sra outdir/SRR12118866.sra --output-format-possibilities fastq --force --output-directory outdir'.format(
                kingfisher
                )],
                stderr=subprocess.PIPE,
                check=True)
            self.assertTrue(os.path.getsize('outdir/SRR12118866_1.fastq')==21411192)
            self.assertTrue(os.path.getsize('outdir/SRR12118866_2.fastq')==21411192)
            self.assertFalse('SRR12118866 as an output file already appears to exist' in r.stderr.decode())

    def test_download_fastq_no_force(self):
        with in_tempdir():
            extern.run('touch SRR12118866.fastq SRR12118866_1.fastq SRR12118866_2.fastq'.format(kingfisher))
            r = subprocess.run(['bash','-c','{} get -m prefetch -r SRR12118866 --output-format-possibilities fastq'.format(
                kingfisher
                )],
                stderr=subprocess.PIPE,
                check=True)
            self.assertTrue('SRR12118866 as an output file already appears to exist' in r.stderr.decode())

    def download_force(self):
        with in_tempdir():
            extern.run('touch SRR12118866.fastq SRR12118866_1.fastq SRR12118866_2.fastq'.format(kingfisher))
            r = subprocess.run(['bash','-c','{} get -m prefetch -r SRR12118866 --output-format-possibilities fastq --force'.format(
                kingfisher
                )],
                stderr=subprocess.PIPE,
                check=True)
            self.assertTrue(os.path.getsize('SRR12118866_1.fastq')==21411192)
            self.assertTrue(os.path.getsize('SRR12118866_2.fastq')==21411192)
            self.assertFalse('SRR12118866 as an output file already appears to exist' in r.stderr.decode())

    def test_aws_failure_curl(self):
        with in_tempdir():
            with self.assertRaises(extern.ExternCalledProcessError):
                extern.run('{} get -r DRR014182_not --force -f sra -m aws-http --guess-aws-location'.format(kingfisher))
            self.assertFalse(os.path.exists('DRR014182.sra'))

    def test_aws_failure_aria2(self):
        with in_tempdir():
            with self.assertRaises(extern.ExternCalledProcessError):
                # As of writing DRR014182 actually doesn't exist at AWS.
                extern.run('{} get -r DRR014182 --force -f sra -m aws-http --guess-aws-location --download-threads 5'.format(kingfisher))
            self.assertFalse(os.path.exists('DRR014182.sra'))

    def test_prefetch_max_size_limit(self):
        with in_tempdir():
            # Prefetch used to have a non-zero exit code when the file was too big, but now it doesn't. Handle both cases.
            try:
                extern.run('{} get -r SRR12118866 -f sra -m prefetch --prefetch-max-size 1M'.format(kingfisher))
            except extern.ExternCalledProcessError as e:
                pass
            self.assertFalse(os.path.exists('DRR014182.sra'))

    def test_get_run_identifiers_list(self):
        with in_tempdir():
            with open('runlist','w') as f:
                f.write('SRR12118864\n')
                f.write('SRR12118866\n')
            extern.run('{} get --run-identifiers-list runlist -f sra -m prefetch'.format(kingfisher))
            self.assertTrue(os.path.exists('SRR12118864.sra'))
            self.assertTrue(os.path.exists('SRR12118866.sra'))

    def test_get_run_identifiers_list_output_directory(self):
        with in_tempdir():
            with open('runlist','w') as f:
                f.write('SRR12118864\n')
                f.write('SRR12118866\n')
            extern.run('{} get --run-identifiers-list runlist -f sra -m prefetch --output-directory outdir'.format(kingfisher))
            self.assertTrue(os.path.exists('outdir/SRR12118864.sra'))
            self.assertTrue(os.path.exists('outdir/SRR12118866.sra'))

    @pytest.mark.aws_cp
    def test_aws_cp(self):
        with in_tempdir():
            extern.run('{} get -r SRR12118866 --force -f sra -m aws-cp'.format(kingfisher))
            self.assertTrue(os.path.exists('SRR12118866.sra'))     
            self.assertTrue(os.path.getsize('SRR12118866.sra')==11643188)

    @pytest.mark.aws_cp
    def test_aws_cp_output_directory(self):
        with in_tempdir():
            extern.run('{} get -r SRR12118866 --force -f sra -m aws-cp --output-directory outdir'.format(kingfisher))
            self.assertTrue(os.path.exists('outdir/SRR12118866.sra'))     
            self.assertTrue(os.path.getsize('outdir/SRR12118866.sra')==11643188)

    def test_aws_http_md5sums(self):
        with in_tempdir():
            extern.run('{} get -r SRR12118866 --force -f sra -m aws-http --check-md5sums 2>kingfisher_stderr'.format(kingfisher))
            self.assertTrue(os.path.exists('SRR12118866.sra'))
            self.assertTrue(os.path.getsize('SRR12118866.sra')==11643188)
            with open('kingfisher_stderr') as f:
                stderr = f.read()
                self.assertTrue(f'MD5sum OK for {os.getcwd()}/SRR12118866.sra' in stderr)

    def test_aws_http_md5sums_output_directory(self):
        with in_tempdir():
            extern.run('{} get -r SRR12118866 --force -f sra -m aws-http --output-directory outdir --check-md5sums 2>kingfisher_stderr'.format(kingfisher))
            self.assertTrue(os.path.exists('outdir/SRR12118866.sra'))
            self.assertTrue(os.path.getsize('outdir/SRR12118866.sra')==11643188)
            with open('kingfisher_stderr') as f:
                stderr = f.read()
                self.assertTrue(f'MD5sum OK for {os.getcwd()}/outdir/SRR12118866.sra' in stderr)

    def test_get_by_bioproject(self):
        with in_tempdir():
            try:
                extern.run('{} get --bioproject PRJNA177893 --force -f sra -m aws-http --check-md5sums 2>kingfisher_stderr'.format(kingfisher))
            except extern.ExternCalledProcessError:
                with open('kingfisher_stderr') as f:
                    raise Exception("kingfisher_stderr:\n" + f.read())
            # -rw-rw-r-- 1 ben ben 45099009 Sep 16 08:01 SRR600121.sra
            # -rw-rw-r-- 1 ben ben 31203681 Sep 16 08:01 SRR605326.sra
            # -rw-rw-r-- 1 ben ben  4291189 Sep 16 08:01 SRR605331.sra
            # -rw-rw-r-- 1 ben ben  4645669 Sep 16 08:01 SRR609442.sra
            # -rw-rw-r-- 1 ben ben 31203681 Sep 16 08:02 SRR609443.sra
            # -rw-rw-r-- 1 ben ben 33424541 Sep 16 08:02 SRR616039.sra
            # -rw-rw-r-- 1 ben ben  7736661 Sep 16 08:02 SRR616042.sra
            # -rw-rw-r-- 1 ben ben  5218613 Sep 16 08:02 SRR616043.sra
            # -rw-rw-r-- 1 ben ben  1468505 Sep 16 08:02 SRR616044.sra
            self.assertTrue(os.path.getsize('SRR600121.sra')==45099009)
            self.assertTrue(os.path.getsize('SRR605326.sra')==31203681)
            self.assertTrue(os.path.getsize('SRR605331.sra')==4291189)
            self.assertTrue(os.path.getsize('SRR609442.sra')==4645669)
            self.assertTrue(os.path.getsize('SRR609443.sra')==31203681)
            self.assertTrue(os.path.getsize('SRR616039.sra')==33424541)
            self.assertTrue(os.path.getsize('SRR616042.sra')==7736661)
            self.assertTrue(os.path.getsize('SRR616043.sra')==5218613)
            self.assertTrue(os.path.getsize('SRR616044.sra')==1468505)

            with open('kingfisher_stderr') as f:
                stderr = f.read()
                self.assertTrue(f'MD5sum OK for {os.getcwd()}/SRR600121.sra' in stderr)

    def test_get_by_bioproject_output_directory(self):
        with in_tempdir():
            extern.run('{} get --bioproject PRJNA177893 --force -f sra -m aws-http --output-directory outdir --check-md5sums 2>kingfisher_stderr'.format(kingfisher))
            # -rw-rw-r-- 1 ben ben 45099009 Sep 16 08:01 SRR600121.sra
            # -rw-rw-r-- 1 ben ben 31203681 Sep 16 08:01 SRR605326.sra
            # -rw-rw-r-- 1 ben ben  4291189 Sep 16 08:01 SRR605331.sra
            # -rw-rw-r-- 1 ben ben  4645669 Sep 16 08:01 SRR609442.sra
            # -rw-rw-r-- 1 ben ben 31203681 Sep 16 08:02 SRR609443.sra
            # -rw-rw-r-- 1 ben ben 33424541 Sep 16 08:02 SRR616039.sra
            # -rw-rw-r-- 1 ben ben  7736661 Sep 16 08:02 SRR616042.sra
            # -rw-rw-r-- 1 ben ben  5218613 Sep 16 08:02 SRR616043.sra
            # -rw-rw-r-- 1 ben ben  1468505 Sep 16 08:02 SRR616044.sra
            self.assertTrue(os.path.getsize('outdir/SRR600121.sra')==45099009)
            self.assertTrue(os.path.getsize('outdir/SRR605326.sra')==31203681)
            self.assertTrue(os.path.getsize('outdir/SRR605331.sra')==4291189)
            self.assertTrue(os.path.getsize('outdir/SRR609442.sra')==4645669)
            self.assertTrue(os.path.getsize('outdir/SRR609443.sra')==31203681)
            self.assertTrue(os.path.getsize('outdir/SRR616039.sra')==33424541)
            self.assertTrue(os.path.getsize('outdir/SRR616042.sra')==7736661)
            self.assertTrue(os.path.getsize('outdir/SRR616043.sra')==5218613)
            self.assertTrue(os.path.getsize('outdir/SRR616044.sra')==1468505)

            with open('kingfisher_stderr') as f:
                stderr = f.read()
                self.assertTrue(f'MD5sum OK for {os.getcwd()}/outdir/SRR600121.sra' in stderr)
    
    def test_api_stability(self):
        from kingfisher import download_and_extract
        with in_tempdir():
            # This is the annotate version of the usage, but testing along the same lines
            # https://github.com/NIAID-Data-Ecosystem/nde-crawlers/blob/c67c77d0c547466cc2d2264f8e53e0c809b9dfd0/ncbi_sra/files/ncbi_sra.py#L34
            download_and_extract(
                run_identifiers=None,
                bioproject_accession='PRJNA177893',
                run_identifiers_file=None,
                download_methods=['aws-http'],
                output_format_possibilities=['sra'],
            )
            self.assertEqual(os.path.getsize('SRR616044.sra'), 1468505)

    # def test_noqual(self):
    #     with in_tempdir():
    #         extern.run("{} -r ERR3209781 --allowable-output-formats ".format(kingfisher, ))
    #         self.assertTrue(os.path.getsize('ERR3209781_1.fasta')==21411192)
    #         self.assertTrue(os.path.getsize('ERR3209781_2.fasta')==21411192)

    # e.g. DRR014182
    # ERR1877729
    # ERR3209781 => all at gcp only, so leave that be for now I think.

if __name__ == "__main__":
    unittest.main()
